import numpy as np
from abc import abstractclassmethod, ABCMeta

from pykin.utils.log_utils import create_logger
from pykin.utils.error_utils import CollisionError, NotFoundError

logger = create_logger('Cartesian Planner', "debug",)

class Planner(metaclass=ABCMeta):
    """
    Base Planner class 

    Args:
        robot (SingleArm or Bimanual): The manipulator robot type is SingleArm or Bimanual
        self_collision_manager: CollisionManager for robot's self collision check
        obstacle_collision_manager: CollisionManager for collision check between robot and object
        dimension(int): robot arm's dof
    """
    def __init__(
        self,
        robot,
        self_collision_manager,
        obstacle_collision_manager,
        dimension
    ):
        self.robot = robot
        self._dimension = dimension
        if self_collision_manager is None:
            logger.warning(f"This Planner does not do collision checking")
            self.self_c_manager = None
        else:
            self.self_c_manager = self_collision_manager
            check_collision = self.self_c_manager.in_collision_internal()
            if check_collision:
                raise CollisionError("Conflict confirmed. Check the joint settings again")
        self.obstacle_c_manager = obstacle_collision_manager

    def __repr__(self) -> str:
        return 'pykin.planners.planner.{}()'.format(type(self).__name__)

    @staticmethod
    def _change_types(datas):
        """
        """
        if not isinstance(datas, (np.ndarray)):
            datas = np.array(datas)
            if datas.size == 0:
                raise NotFoundError("Make sure set current or goal joints..")
        return datas

    def _setup_q_limits(self):
        """
        Setup joint limits (lower and upper)
        """
        if self.arm is not None:
            self.q_limits_lower = self.robot.joint_limits_lower[self.arm]
            self.q_limits_upper = self.robot.joint_limits_upper[self.arm]
        else:
            self.q_limits_lower = self.robot.joint_limits_lower
            self.q_limits_upper = self.robot.joint_limits_upper

    def _check_q_in_limits(self, q_in):
        """
        check q_in within joint limits
        If q_in is in joint limits, return True
        otherwise, return False

        Returns:
            bool(True or False)
        """
        return np.all([q_in >= self.q_limits_lower, q_in <= self.q_limits_upper])

    def _setup_eef_name(self):
        """
        Setup end-effector name
        """
        if self.arm is not None:
            self.eef_name = self.robot.eef_name[self.arm]

    def _collision_free(self, new_q, visible_name=False):
        """
        Check collision free between robot and obstacles
        If visible name is True, return collision result and collision object names
        otherwise, return only collision result

        Args:
            new_q(np.array): new joint angles
            visible_name(bool)

        Returns:
            result(bool): If collision free, return True
            names(set of 2-tup): The set of pairwise collisions. 
        """
 
        if self.self_c_manager is None:
            return True

        transformations = self._get_transformations(new_q)

        for link, transformations in transformations.items():
            if link in self.self_c_manager._objs:
                transform = transformations.h_mat
                A2B = np.dot(transform, self.robot.links[link].visual.offset.h_mat)
                self.self_c_manager.set_transform(name=link, transform=A2B)
        
        is_self_collision = self.self_c_manager.in_collision_internal(return_names=False, return_data=False)
        is_obstacle_collision = self.self_c_manager.in_collision_other(other_manager=self.obstacle_c_manager, return_names=False)  

        name = None
        if visible_name:
            if is_self_collision:
                return False, name
            return True, name

        if is_self_collision or is_obstacle_collision:
            return False
        return True

    @abstractclassmethod
    def get_path_in_joinst_space(self):
        """
        write planner algorithm you want 
        """
        raise NotImplementedError

    @abstractclassmethod
    def _get_linear_path(self, init_pose, goal_pose):
        raise NotImplementedError

    def _get_transformations(self, q_in):
        """
        Get transformations corresponding to q_in

        Args:
            q_in(np.array): joint angles

        Returns:
            transformations(OrderedDict)
        """
        if self.robot.robot_name == "sawyer":
            q_in = np.concatenate((np.zeros(1), q_in))

        if self.arm is not None:
            transformations = self.robot.forward_kin(q_in, self.robot.desired_frames[self.arm])
        else:
            transformations = self.robot.forward_kin(q_in)
        return transformations

    @property
    def dimension(self):
        return self._dimension

    @dimension.setter
    def dimension(self, dimesion):
        self._dimension = dimesion

    @property
    def cur_qpos(self):
        return self._cur_qpos

    @cur_qpos.setter
    def cur_qpos(self, cur_qpos):
        self._cur_qpos = cur_qpos

    @property
    def goal_pose(self):
        return self._goal_pose

    @goal_pose.setter
    def goal_pose(self, goal_pose):
        self._goal_pose = goal_pose


import sys, os
import numpy as np

import signal
def handler(signum, frame):
    exit()
# Set the signal handler
signal.signal(signal.SIGINT, handler)

pykin_path = os.path.abspath(os.path.dirname(__file__)+"../" )
sys.path.append(pykin_path)

from pykin.kinematics.transform import Transform
from pykin.kinematics.kinematics import Kinematics
from pykin.models.urdf_model import URDFModel
from pykin.utils.transform_utils import compute_pose_error

class Robot(URDFModel):
    """
    Initializes a robot object, as defined by a single corresponding robot URDF

    Args:
        fname (str): path to the urdf file.
        offset (Transform): robot init offset
    """
    def __init__(
        self, 
        fname=None, 
        offset=None, 
    ):
        if fname is None:
            fname = pykin_path + "/asset/urdf/baxter/baxter.urdf"
        self._offset = offset
        if offset is None:
            self._offset = Transform()
            
        super(Robot, self).__init__(fname)

        self.joint_limits_lower = []
        self.joint_limits_upper = []

        self._setup_kinematics()
        self._setup_init_transform()

        self.joint_limits = self._get_limited_joint_names()

    def __str__(self):
        return f"""ROBOT : {self.robot_name} 
        {self.links} 
        {self.joints}"""

    def __repr__(self):
        return 'pykin.robot.{}()'.format(type(self).__name__)

    def show_robot_info(self):
        """
        Shows robot's info 
        """
        print("*" * 100)
        print(f"Robot Information:")

        for link in self.links.values():
            print(link)
        for joint in self.joints.values():
            print(joint)
            
        print(f"robot's dof : {self.dof}")
        print(f"active joint names: \n{self.get_all_active_joint_names()}")
        print(f"revolute joint names: \n{self.get_revolute_joint_names()}")
        print("*" * 100)

    def _setup_kinematics(self):
        """
        Setup Kinematics
        """
        self.kin = Kinematics(robot_name=self.robot_name,
                              offset=self.offset,
                              active_joint_names=super().get_revolute_joint_names(),
                              base_name="", 
                              eef_name=None
                              )
    
    def _setup_init_transform(self):
        """
        Initializes robot's transformation
        """
        thetas = np.zeros(len(super().get_revolute_joint_names()))
        transformations = self.kin.forward_kinematics(self.root, thetas)
        self.init_transformations = transformations

    def _get_limited_joint_names(self):
        result = {}
        for joint, value in self.joints.items():
            for active_joint in super().get_revolute_joint_names():
                if joint == active_joint:
                    result.update({joint : (value.limit[0], value.limit[1])})
        return result

    def setup_link_name(self, base_name, eef_name):
        """
        Sets robot's link name

        Args:
            base_name (str): reference link name
            eef_name (str): end effector name
        """
        raise NotImplementedError

    def forward_kin(self, thetas):
        self._frames = self.root
        transformation = self.kin.forward_kinematics(self._frames, thetas)
        return transformation

    def inverse_kin(self, current_joints, target_pose, method, maxIter):
        raise NotImplementedError

    def _set_joint_limits_upper_and_lower(self):
        raise NotImplementedError

    def get_pose_error(self, target=np.eye(4), result=np.eye(4)):
        return compute_pose_error(target, result)

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, offset):
        self._offset = offset

    @property
    def base_name(self):
        raise NotImplementedError
    
    @property
    def eef_name(self):
        raise NotImplementedError

    @property
    def frame(self):
        raise NotImplementedError

    @property
    def active_joint_names(self):
        raise NotImplementedError




import numpy as np
import trimesh
from abc import abstractclassmethod

# import pykin.utils.pnp_utils

import pykin.utils.plot_utils as plt
from pykin.collision.collision_manager import CollisionManager
from pykin.utils.task_utils import get_transform


class ActivityBase:
    def __init__(
        self,
        robot,
        robot_col_manager,
        obstacles_col_manager,
        mesh_path,
        **gripper_configures
    ):
        self.robot = robot
        self.robot_c_manager = robot_col_manager
        self.obstacles_c_manager = obstacles_col_manager
        self.gripper_c_manager = CollisionManager()
        self.mesh_path = mesh_path
        self.gripper_names = gripper_configures.get("gripper_names", None)
        self.gripper_names.insert(0, self.robot.eef_name)
        self.gripper_max_width = gripper_configures.get("gripper_max_width", 0.0)
        self.gripper_max_depth = gripper_configures.get("gripper_max_depth", 0.0)
        self.tcp_position = gripper_configures.get("tcp_position", np.zeros(3))
        self.gripper = self._generate_gripper()

    def __repr__(self) -> str:
        return 'pykin.tasks.activity.{}()'.format(type(self).__name__)

    @abstractclassmethod
    def generate_tcp_poses(self):
        pass

    def _generate_gripper(self):
        gripper = {}
        for link, transform in self.robot.init_transformations.items():
            if link in self.gripper_names:
                gripper[link] = transform.h_mat
                if self.robot.links[link].collision.gtype == "mesh":
                    mesh_name = self.robot.links[link].collision.gparam.get('filename')
                    file_name = self.mesh_path + mesh_name
                    mesh = trimesh.load_mesh(file_name)
                    A2B = np.dot(transform.h_mat, self.robot.links[link].collision.offset.h_mat)
                    self.gripper_c_manager.add_object(link, gtype="mesh", gparam=mesh, transform=A2B)
        return gripper

    def get_gripper(self):
        return self.gripper

    def get_gripper_transformed(self, pose, is_tcp=True):
        transform_gripper = {}
        gripper = self.get_gripper()
        tcp_pose = pose

        if not is_tcp:
            tcp_pose = self.get_tcp_h_mat_from_eef(pose)

        for link, transform in gripper.items():
            T = get_transform(gripper[self.gripper_names[-1]], tcp_pose)
            transform_gripper[link] = np.dot(T, transform)
        return transform_gripper

    def collision_free(self, transformations, only_gripper=False) -> bool:
        for link, transform in transformations.items():
            if only_gripper:
                if self.robot.links[link].collision.gtype == "mesh":
                    A2B = np.dot(transform, self.robot.links[link].collision.offset.h_mat)
                    self.gripper_c_manager.set_transform(link, A2B)
            else:
                for link, transform in transformations.items():
                    if link in self.robot_c_manager._objs:
                        A2B = np.dot(transform.h_mat, self.robot.links[link].collision.offset.h_mat)
                        self.robot_c_manager.set_transform(name=link, transform=A2B)
          
        if only_gripper:
            is_obstacle_collision = self.gripper_c_manager.in_collision_other(other_manager=self.obstacles_c_manager)
            if is_obstacle_collision:
                return False
            return True
        else:
            is_self_collision = self.robot_c_manager.in_collision_internal()
            is_obstacle_collision = self.robot_c_manager.in_collision_other(other_manager=self.obstacles_c_manager)
            if is_self_collision or is_obstacle_collision:
                return False
            return True

    def get_tcp_pose(self, transformations):
        return transformations["tcp"]

    def get_eef_h_mat_from_tcp(self, tcp_pose):
        eef_pose = np.eye(4)
        eef_pose[:3, :3] = tcp_pose[:3, :3]
        eef_pose[:3, 3] = tcp_pose[:3, 3] - np.dot(self.tcp_position[-1], tcp_pose[:3, 2])
        return eef_pose

    def get_tcp_h_mat_from_eef(self, eef_pose):
        tcp_pose = np.eye(4)
        tcp_pose[:3, :3] = eef_pose[:3, :3]
        tcp_pose[:3, 3] = eef_pose[:3, 3] + np.dot(self.tcp_position[-1], eef_pose[:3, 2])
        return tcp_pose

    def visualize_robot(
        self, 
        ax, 
        transformations, 
        alpha=1.0,
        only_gripper=False
    ):
        plt.plot_basis(self.robot, ax)
        for link, transform in transformations.items():
            if "pedestal" in link:
                continue
            if self.robot.links[link].collision.gtype == "mesh":
                mesh_name = self.mesh_path + self.robot.links[link].collision.gparam.get('filename')
                mesh = trimesh.load_mesh(mesh_name)
                A2B = np.dot(transform.h_mat, self.robot.links[link].collision.offset.h_mat)
                color = self.robot.links[link].collision.gparam.get('color')
                color = np.array([color for color in color.values()]).flatten()
                mesh.visual.face_colors = color
                
                if only_gripper:
                    if link in self.gripper_names:
                        plt.plot_mesh(ax=ax, mesh=mesh, A2B=A2B, alpha=alpha, color=color)
                else:
                    plt.plot_mesh(ax=ax, mesh=mesh, A2B=A2B, alpha=alpha, color=color)

    def visualize_gripper(
        self,
        ax,
        gripper,
        alpha=1.0,
        color=None,
        visible_basis=False
    ):
        if visible_basis:
            plt.plot_basis(self.robot, ax)
        for link, transform in gripper.items():
            if self.robot.links[link].collision.gtype == "mesh":
                mesh_name = self.mesh_path + self.robot.links[link].collision.gparam.get('filename')
                mesh = trimesh.load_mesh(mesh_name)
                A2B = np.dot(transform, self.robot.links[link].collision.offset.h_mat)

                mesh_color = color
                if color is None:
                    mesh_color = self.robot.links[link].collision.gparam.get('color')
                    mesh_color = np.array([color for color in mesh_color.values()]).flatten()
                if "finger" in link:
                    alpha = 1
                plt.plot_mesh(ax=ax, mesh=mesh, A2B=A2B, alpha=alpha, color=mesh_color)

    def visualize_axis(
        self,
        ax,
        transformation,
        link=None,
        axis=[1, 1, 1],
        scale=0.1
        
    ):
        pose = transformation
        if link is not None:
            pose = transformation[link].h_mat
        if axis[0]:
            plt.plot_normal_vector(ax, pose[:3, 3], pose[:3, 0], scale=scale, edgecolor="red")
        if axis[1]:
            plt.plot_normal_vector(ax, pose[:3, 3], pose[:3, 1], scale=scale, edgecolor="green")
        if axis[2]:
            plt.plot_normal_vector(ax, pose[:3, 3], pose[:3, 2], scale=scale, edgecolor="blue")

    def visualize_point(
        self,
        ax,
        transformation,
        link=None
    ):
        pose = transformation
        if link is not None:
            pose = transformation[link].h_mat

        plt.plot_basis(self.robot, ax)
        plt.plot_vertices(ax, pose[:3, 3])   


import numpy as np
from collections import OrderedDict

from pykin.kinematics import jacobian as jac
from pykin.utils import transform_utils as tf
from pykin.utils.kin_utils import Baxter, calc_pose_error, convert_thetas_to_dict, logging_time

class Kinematics:
    """
    Class of Kinematics

    Args:
        robot_name (str): robot's name
        offset (Transform): robot's offset
        active_joint_names (list): robot's actuated joints
        base_name (str): reference link's name
        eef_name (str): end effector's name
    """
    def __init__(self, 
                robot_name, 
                offset, 
                active_joint_names=[],
                base_name="base", 
                eef_name=None, 
                ):
        self.robot_name = robot_name
        self.offset = offset
        self.active_joint_names = active_joint_names
        self.base_name = base_name
        self.eef_name = eef_name

    def forward_kinematics(self, frames, thetas):
        """
        Returns transformations obtained by computing fk

        Args:
            frames (list or Frame()): robot's frame for forward kinematics
            thetas (sequence of float): input joint angles

        Returns:
            transformations (OrderedDict): transformations
        """

        if not isinstance(frames, (list, dict)) :
            thetas = convert_thetas_to_dict(self.active_joint_names, thetas)
        transformations = self._compute_FK(frames, self.offset, thetas)
        return transformations
    
    @logging_time
    def inverse_kinematics(self, frames, current_joints, target_pose, method="LM", maxIter=1000):
        """
        Returns joint angles obtained by computing IK
        
        Args:
            frames (Frame()): robot's frame for invers kinematics
            current_joints (sequence of float): input joint angles
            target_pose (np.array): goal pose to achieve
            method (str): two methods to calculate IK (LM: Levenberg-marquardt, NR: Newton-raphson)
            maxIter (int): Maximum number of calculation iterations

        Returns:
            joints (np.array): target joint angles
        """
        if method == "NR":
            joints = self._compute_IK_NR(
                frames,
                current_joints, 
                target_pose, 
                maxIter=maxIter
            )
        if method == "LM":
            joints = self._compute_IK_LM(
                frames,
                current_joints, 
                target_pose, 
                maxIter=maxIter
            )
        return joints

    def _compute_FK(self, frames, offset, thetas):
        """
        Computes forward kinematics

        Args:
            frames (list or Frame()): robot's frame for forward kinematics
            offset (Transform): robot's offset
            thetas (sequence of float): input joint angles

        Returns:
            transformations (OrderedDict): transformations
        """
        transformations = OrderedDict()
        if not isinstance(frames, list):
            trans = offset * frames.get_transform(thetas.get(frames.joint.name, 0.0))
            transformations[frames.link.name] = trans * frames.link.offset
            for child in frames.children:
                transformations.update(self._compute_FK(child, trans, thetas))
        else:
            # To compute IK
            cnt = 0
            trans = offset
            for frame in frames:
                trans = trans * frame.get_transform(thetas[cnt])
                transformations[frame.link.name] = trans * frame.link.offset
                
                if frame.joint.dtype != "fixed":
                    cnt += 1
                
                if cnt >= len(thetas):
                    cnt -= 1     
                
                if self.robot_name == "baxter":
                    Baxter.add_visual_link(transformations, frame)

        return transformations

    def _compute_IK_NR(self, frames, current_joints, target_pose, maxIter):
        """
        Computes inverse kinematics using Newton Raphson method

        Args:
            frames (list or Frame()): robot's frame for inverse kinematics
            current_joints (sequence of float): input joint angles
            target_pose (np.array): goal pose to achieve
            maxIter (int): Maximum number of calculation iterations

        Returns:
            joints (np.array): target joint angles
        """
        lamb = 0.5
        iterator = 1
        EPS = float(1e-6)
        dof = len(current_joints)

        target_pose = tf.get_h_mat(target_pose[:3], target_pose[3:])

        cur_fk = self.forward_kinematics(frames, current_joints)
        cur_pose = list(cur_fk.values())[-1].h_mat

        err_pose = calc_pose_error(target_pose, cur_pose, EPS)
        err = np.linalg.norm(err_pose)

        while err > EPS:

            iterator += 1
            if iterator > maxIter:
                break
            
            J = jac.calc_jacobian(frames, cur_fk, len(current_joints))
            dq = lamb * np.dot(np.linalg.pinv(J), err_pose)
            current_joints = [current_joints[i] + dq[i] for i in range(dof)]
            cur_fk = self.forward_kinematics(frames, current_joints)

            cur_pose = list(cur_fk.values())[-1].h_mat
            err_pose = calc_pose_error(target_pose, cur_pose, EPS)
            err = np.linalg.norm(err_pose)

        print(f"Iterators : {iterator-1}")
        current_joints = np.array([float(current_joint) for current_joint in current_joints])
        return current_joints

    def _compute_IK_LM(self, frames, current_joints, target, maxIter):
        """
        Computes inverse kinematics using Levenberg-Marquatdt method

        Args:
            frames (list or Frame()): robot's frame for inverse kinematics
            current_joints (sequence of float): input joint angles
            target_pose (np.array): goal pose to achieve
            maxIter (int): Maximum number of calculation iterations

        Returns:
            joints (np.array): target joint angles
        """
        iterator = 1
        EPS = float(1E-12)
        dof = len(current_joints)
        wn_pos = 1/0.3
        wn_ang = 1/(2*np.pi)
        We = np.diag([wn_pos, wn_pos, wn_pos, wn_ang, wn_ang, wn_ang])
        Wn = np.eye(dof)

        target_pose = tf.get_h_mat(target[:3], target[3:])

        cur_fk = self.forward_kinematics(frames, current_joints)
        cur_pose = list(cur_fk.values())[-1].h_mat

        err = calc_pose_error(target_pose, cur_pose, EPS)
        Ek = float(np.dot(np.dot(err.T, We), err)[0])

        while Ek > EPS:
            iterator += 1
            if iterator > maxIter:
                break
            
            lamb = Ek + 0.002

            J = jac.calc_jacobian(frames, cur_fk, len(current_joints))
            J_dls = np.dot(np.dot(J.T, We), J) + np.dot(Wn, lamb)
            
            gerr = np.dot(np.dot(J.T, We), err)
            dq = np.dot(np.linalg.inv(J_dls), gerr)
            current_joints = [current_joints[i] + dq[i] for i in range(dof)]
           
            cur_fk = self.forward_kinematics(frames, current_joints)
            cur_pose = list(cur_fk.values())[-1].h_mat
            err = calc_pose_error(target_pose, cur_pose, EPS)
            Ek2 = float(np.dot(np.dot(err.T, We), err)[0])
            
            if Ek2 < Ek:
                Ek = Ek2
            else:
                current_joints = [current_joints[i] - dq[i] for i in range(dof)]
                cur_fk = self.forward_kinematics(frames, current_joints)
                break
            
        print(f"Iterators : {iterator-1}")
        current_joints = np.array([float(current_joint) for current_joint in current_joints])
        return current_joints


from pykin.kinematics.transform import Transform


class Visual:
    """
    class of Visual

    Args:
        offset (Transform): visual offset
        geom_type (str): visual type (box, cylinder, spehre, mesh)
        geom_param (dict): visual parameters 
    """
    TYPES = ['box', 'cylinder', 'sphere', 'mesh']
    def __init__(
        self, 
        offset=Transform(), 
        geom_type=None, 
        geom_param=None
    ):
        self.offset = offset
        self.gtype = geom_type
        self.gparam = geom_param

    def __str__(self):
        return f"""Visual(offset={self.offset},
                           geom_type={self.gtype}, 
                           geom_param={self.gparam})"""

    def __repr__(self):
        return 'pykin.geometry.geometry.{}()'.format(type(self).__name__)

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, offset):
        self._offset = Transform(offset.pos, offset.rot)


class Collision:
    """
    class of Collision

    Args:
        offset (Transform): collision offset
        geom_type (str): collision type (box, cylinder, spehre, mesh)
        geom_param (dict): collision parameters 
    """
    TYPES = ['box', 'cylinder', 'sphere', 'mesh']
    def __init__(
        self, 
        offset=Transform(), 
        geom_type=None, 
        geom_param=None
    ):
        self.offset = offset
        self.gtype = geom_type
        self.gparam = geom_param

    def __str__(self):
        return f"""Collision(offset={self.offset},
                              geom_type={self.gtype}, 
                              geom_param={self.gparam})"""

    def __repr__(self):
        return 'pykin.geometry.geometry.{}()'.format(type(self).__name__)

    @property
    def offset(self):
        return self._offset

    @offset.setter
    def offset(self, offset):
        self._offset = Transform(offset.pos, offset.rot)


import numpy as np
from enum import Enum, auto
from collections import OrderedDict
from copy import deepcopy

from pykin.tasks.activity import ActivityBase
from pykin.utils.task_utils import normalize, surface_sampling, projection, get_rotation_from_vectors
from pykin.utils.transform_utils import get_pose_from_homogeneous
from pykin.utils.log_utils import create_logger

logger = create_logger('Grasp', "debug")

class GraspStatus(Enum):
    pre_grasp_pose = auto()
    grasp_pose = auto()
    post_grasp_pose = auto()
    pre_release_pose = auto()
    release_pose = auto()
    post_release_pose = auto()

class GraspManager(ActivityBase):
    def __init__(
        self,
        robot,
        robot_col_manager,
        obstacles_col_manager,
        mesh_path,
        retreat_distance = 0.1,
        release_distance = 0.01,
        **gripper_configures
    ):
        super().__init__(
            robot,
            robot_col_manager,
            obstacles_col_manager,
            mesh_path,
            **gripper_configures)

        self.retreat_distance = retreat_distance
        self.release_distance = release_distance
        self.tcp_pose = np.eye(4)
        self.result_obj_pose = np.eye(4)
        self.contact_points = None

    def get_grasp_waypoints(
        self,
        obj_mesh,
        obj_pose,
        limit_angle,
        num_grasp=1,
        n_trials=1,
    ):
        waypoints = OrderedDict()

        grasp_pose = self.get_grasp_pose(obj_mesh, obj_pose, limit_angle, num_grasp, n_trials)    
        waypoints[GraspStatus.pre_grasp_pose] = self.pre_grasp_pose
        waypoints[GraspStatus.grasp_pose] = grasp_pose
        waypoints[GraspStatus.post_grasp_pose] =self.post_grasp_pose

        return waypoints
        
    def get_grasp_pose(        
        self,
        obj_mesh,
        obj_pose,
        limit_angle,
        num_grasp=1,
        n_trials=1,
    ):
        grasp_poses = self.generate_grasps(obj_mesh, obj_pose, limit_angle, num_grasp, n_trials)
        grasp_pose = self.filter_grasps(grasp_poses)     
        return grasp_pose

    def get_pre_grasp_pose(self, grasp_pose):
        pre_grasp_pose = np.eye(4)
        pre_grasp_pose[:3, :3] = grasp_pose[:3, :3]
        pre_grasp_pose[:3, 3] = grasp_pose[:3, 3] - self.retreat_distance * grasp_pose[:3,2]    
        # pre_grasp_pose[:3, 3] = grasp_pose[:3, 3] + np.array([0, 0, self.retreat_distance])  
        return pre_grasp_pose

    def get_post_grasp_pose(self, grasp_pose):
        post_grasp_pose = np.eye(4)
        post_grasp_pose[:3, :3] = grasp_pose[:3, :3] 
        post_grasp_pose[:3, 3] = grasp_pose[:3, 3] + np.array([0, 0, self.retreat_distance])  
        return post_grasp_pose

    def generate_grasps(
        self,
        obj_mesh,
        obj_pose,
        limit_angle=0.05, #radian
        num_grasp=1,
        n_trials=1
    ):
        cnt = 0
        while cnt < num_grasp * n_trials:
            tcp_poses = self.generate_tcp_poses(obj_mesh, obj_pose, limit_angle, n_trials)
            for tcp_pose, contact_points, _ in tcp_poses:
                gripper_transformed = self.get_gripper_transformed(tcp_pose)

                if self.collision_free(gripper_transformed, only_gripper=True):
                    self.tcp_pose = tcp_pose
                    self.contact_points = contact_points
                    eef_pose = self.get_eef_h_mat_from_tcp(tcp_pose)
                    cnt += 1
                    yield (eef_pose, gripper_transformed)
    
    def filter_grasps(self, grasp_poses):
        is_success_filtered = False
        for grasp_pose, _ in grasp_poses:
            qpos = self._compute_inverse_kinematics(grasp_pose)
            if qpos is None:
                continue

            transforms = self.robot.forward_kin(np.array(qpos))
            goal_pose = transforms[self.robot.eef_name].h_mat
 
            if self._check_ik_solution(grasp_pose, goal_pose) and self.collision_free(transforms):
                pre_grasp_pose = self.get_pre_grasp_pose(grasp_pose)
                pre_qpos = self._compute_inverse_kinematics(pre_grasp_pose)
                pre_transforms = self.robot.forward_kin(np.array(pre_qpos))
                pre_goal_pose = pre_transforms[self.robot.eef_name].h_mat

                if self._check_ik_solution(pre_grasp_pose, pre_goal_pose) and self.collision_free(pre_transforms):
                    self.pre_grasp_pose = pre_grasp_pose
                    self.post_grasp_pose = self.get_post_grasp_pose(grasp_pose)
                    is_success_filtered = True
                    break

        if not is_success_filtered:
            logger.error(f"Failed to filter grasp poses")
            return None
        logger.info(f"Success to get grasp pose.\n")

        return grasp_pose

    def generate_tcp_poses(
        self,
        obj_mesh,
        obj_pose,
        limit_angle,
        n_trials
    ):
        contact_points, normals = self._generate_contact_points(obj_mesh, obj_pose, limit_angle)
        p1, p2 = contact_points
        center_point = (p1 + p2) /2
        line = p2 - p1

        for i, grasp_dir in enumerate(self._generate_grasp_directions(line, n_trials)):
            y = normalize(line)
            z = grasp_dir
            x = np.cross(y, z)

            tcp_pose = np.eye(4)
            tcp_pose[:3,0] = x
            tcp_pose[:3,1] = y
            tcp_pose[:3,2] = z
            tcp_pose[:3,3] = center_point

            yield (tcp_pose, contact_points, normals)

    def get_release_waypoints(
        self,
        obj_mesh_on_sup,
        obj_pose_on_sup,
        n_samples_on_sup,
        obj_mesh_for_sup,
        obj_pose_for_sup,
        n_samples_for_sup,
        n_trials=1
    ):
        waypoints = OrderedDict()

        release_pose = self.get_release_pose(
            obj_mesh_on_sup,
            obj_pose_on_sup,
            n_samples_on_sup,
            obj_mesh_for_sup,
            obj_pose_for_sup,
            n_samples_for_sup, 
            n_trials)
        
        waypoints[GraspStatus.pre_release_pose] = self.pre_release_pose
        waypoints[GraspStatus.release_pose] = release_pose
        waypoints[GraspStatus.post_release_pose] =self.post_release_pose

        return waypoints
        
    def get_release_pose(
        self,
        obj_mesh_on_sup,
        obj_pose_on_sup,
        n_samples_on_sup,
        obj_mesh_for_sup,
        obj_pose_for_sup,
        n_samples_for_sup,
        n_trials
    ):
        support_poses = self.generate_supports(
            obj_mesh_on_sup,
            obj_pose_on_sup,
            n_samples_on_sup,
            obj_mesh_for_sup,
            obj_pose_for_sup,
            n_samples_for_sup,
            n_trials)

        release_pose = self.filter_supports(support_poses)
        return release_pose

    def get_pre_release_pose(self, release_pose):
        pre_release_pose = np.eye(4)
        pre_release_pose[:3, :3] = release_pose[:3, :3]
        pre_release_pose[:3, 3] = release_pose[:3, 3] + np.array([0, 0, self.retreat_distance])
        return pre_release_pose

    def get_post_release_pose(self, grasp_pose):
        post_release_pose = np.eye(4)
        post_release_pose[:3, :3] = grasp_pose[:3, :3] 
        post_release_pose[:3, 3] = grasp_pose[:3, 3] - self.retreat_distance * grasp_pose[:3,2]
        return post_release_pose

    def generate_supports(
        self,
        obj_mesh_on_sup,
        obj_pose_on_sup,
        n_samples_on_sup,
        obj_mesh_for_sup,
        obj_pose_for_sup,
        n_samples_for_sup,
        n_trials=1
    ):
        cnt = 0
        self.obj_mesh_for_sup = deepcopy(obj_mesh_for_sup)
        self.obj_mesh_on_sup = deepcopy(obj_mesh_on_sup)
        self.obj_mesh_on_sup.apply_transform(obj_pose_on_sup)
        while cnt < n_trials:
            support_points = self.sample_supports(obj_mesh_on_sup, obj_pose_on_sup, n_samples_on_sup,
                                            obj_mesh_for_sup, obj_pose_for_sup, n_samples_for_sup)
            
            for result_obj_pose, obj_pose_transformed_for_sup, point_on_sup, point_transformed in self._transform_points_on_support(support_points, obj_pose_for_sup):
                T = np.dot(obj_pose_for_sup, np.linalg.inv(obj_pose_transformed_for_sup))
                self.obj_pose_transformed_for_sup = obj_pose_transformed_for_sup
                gripper_pose_transformed = np.dot(T, self.tcp_pose)
                result_gripper_pose = np.eye(4)
                result_gripper_pose[:3, :3] = gripper_pose_transformed[:3, :3]
                result_gripper_pose[:3, 3] = gripper_pose_transformed[:3, 3] + (point_on_sup - point_transformed) + np.array([0, 0, self.release_distance])

                gripper_transformed = self.get_gripper_transformed(result_gripper_pose)
                if self.collision_free(gripper_transformed, only_gripper=True):
                    release_pose = gripper_transformed[self.robot.eef_name]
                    yield release_pose, result_obj_pose
            cnt += 1

    def sample_supports(
        self,
        obj_mesh_on_sup,
        obj_pose_on_sup,
        n_samples_on_sup,
        obj_mesh_for_sup,
        obj_pose_for_sup,
        n_samples_for_sup,
    ):
        sample_points_on_support = self.generate_points_on_support(obj_mesh_on_sup, obj_pose_on_sup, n_samples_on_sup)
        sample_points_for_support = list(self.generate_points_for_support(obj_mesh_for_sup, obj_pose_for_sup, n_samples_for_sup))

        for point_on_support, normal_on_support in sample_points_on_support:
            for point_for_support, normal_for_support in sample_points_for_support:
                yield point_on_support, normal_on_support, point_for_support, normal_for_support

    def _transform_points_on_support(self, support_points, obj_pose_for_sup):
        for point_on_sup, normal_on_sup, point_for_sup, normal_for_sup in support_points:
            normal_on_sup = -normal_on_sup
            R_mat = get_rotation_from_vectors(normal_for_sup, normal_on_sup)
            
            obj_pose_transformed_for_sup = np.eye(4)
            obj_pose_transformed_for_sup[:3, :3] = np.dot(R_mat, obj_pose_for_sup[:3, :3])
            obj_pose_transformed_for_sup[:3, 3] = obj_pose_for_sup[:3, 3]

            point_transformed = np.dot(point_for_sup - obj_pose_for_sup[:3, 3], R_mat) + obj_pose_for_sup[:3, 3]
            # normal_transformed = np.dot(normal_for_sup, R_mat)

            result_obj_pose = np.eye(4)
            result_obj_pose[:3, :3] = obj_pose_transformed_for_sup[:3, :3]
            result_obj_pose[:3, 3] = obj_pose_for_sup[:3, 3] + (point_on_sup - point_transformed) + np.array([0, 0, self.release_distance])

            yield result_obj_pose, obj_pose_transformed_for_sup, point_on_sup, point_transformed

    def filter_supports(self, support_poses):
        is_success_filtered = False
        for release_pose, result_obj_pose in support_poses:
            if not self._check_support(result_obj_pose):
                continue

            qpos = self._compute_inverse_kinematics(release_pose)
            if qpos is None:
                continue

            transforms = self.robot.forward_kin(np.array(qpos))
            goal_pose = transforms[self.robot.eef_name].h_mat

            if self._check_ik_solution(release_pose, goal_pose) and self.collision_free(transforms):
                pre_release_pose = self.get_pre_release_pose(release_pose)
                pre_qpos = self._compute_inverse_kinematics(pre_release_pose)
                pre_transforms = self.robot.forward_kin(np.array(pre_qpos))
                pre_goal_pose = pre_transforms[self.robot.eef_name].h_mat

                if self._check_ik_solution(pre_release_pose, pre_goal_pose) and self.collision_free(pre_transforms):
                    self.pre_release_pose = pre_release_pose
                    # self.post_release_pose = pre_release_pose
                    self.post_release_pose = self.get_post_release_pose(release_pose)
                    self.result_obj_pose = result_obj_pose
                    is_success_filtered = True
                    break

        if not is_success_filtered:
            logger.error(f"Failed to filter release poses")
            return None
        
        logger.info(f"Success to get release pose.\n")
        return release_pose

    def generate_points_on_support(
        self,
        obj_mesh,
        obj_pose,
        n_samples
    ):
        copied_mesh = deepcopy(obj_mesh)
        copied_mesh.apply_transform(obj_pose)

        weights = np.zeros(len(copied_mesh.faces))
        for idx, vertex in enumerate(copied_mesh.vertices[copied_mesh.faces]):
            weights[idx]=0.0
            if np.all(vertex[:,2] >= copied_mesh.bounds[1][2] * 0.98):                
                weights[idx] = 1.0

        support_points, face_ind, normal_vectors = surface_sampling(copied_mesh, n_samples, weights)
        for point, normal_vector in zip(support_points, normal_vectors):
            yield point, normal_vector

    def generate_points_for_support(
        self,
        obj_mesh,
        obj_pose,
        n_samples
    ):
        copied_mesh = deepcopy(obj_mesh)
        copied_mesh.apply_transform(obj_pose)
    
        weights = np.zeros(len(copied_mesh.faces))
        for idx, vertex in enumerate(copied_mesh.vertices[copied_mesh.faces]):
            weights[idx]=0.4
            if np.all(vertex[:,2] <= copied_mesh.bounds[0][2] * 1.02):                
                weights[idx] = 0.6
  
        support_points, face_ind, normal_vectors = surface_sampling(copied_mesh, n_samples, weights)
        for point, normal_vector in zip(support_points, normal_vectors):
            yield point, normal_vector

    def _compute_inverse_kinematics(self, grasp_pose):
        eef_pose = get_pose_from_homogeneous(grasp_pose)
        qpos = self.robot.inverse_kin(np.random.randn(7), eef_pose, maxIter=500)
        return qpos

    def _generate_contact_points(
        self,
        obj_mesh,
        obj_pose,
        limit_angle
    ):
        copied_mesh = deepcopy(obj_mesh)
        copied_mesh.apply_transform(obj_pose)

        while True:
            contact_points, _, normals = surface_sampling(copied_mesh, n_samples=2)
            if self._is_force_closure(contact_points, normals, limit_angle):
                break
        return (contact_points, normals)

    def _is_force_closure(self, vertices, normals, limit_angle):
        vectorA = vertices[0]
        vectorB = vertices[1]

        normalA = -normals[0]
        normalB = -normals[1]

        vectorAB = vectorB - vectorA
        distance = np.linalg.norm(vectorAB)

        unit_vectorAB = normalize(vectorAB)
        angle_A2AB = np.arccos(normalA.dot(unit_vectorAB))

        unit_vectorBA = -1 * unit_vectorAB
        angle_B2AB = np.arccos(normalB.dot(unit_vectorBA))

        if distance > self.gripper_max_width:
            return False

        if angle_A2AB > limit_angle or angle_B2AB > limit_angle:
            return False
        
        return True

    def _generate_grasp_directions(self, vector, n_trials):
        norm_vector = normalize(vector)
        e1, e2 = np.eye(3)[:2]
        v1 = e1 - projection(e1, norm_vector)
        v1 = normalize(v1)
        v2 = e2 - projection(e2, norm_vector) - projection(e2, v1)
        v2 = normalize(v2)

        for theta in np.linspace(-np.pi, np.pi, n_trials):
            normal_dir = np.cos(theta) * v1 + np.sin(theta) * v2
            yield normal_dir

    def _check_ik_solution(self, eef_pose, goal_pose, err_limit=1e-2) -> bool:
        error_pose = self.robot.get_pose_error(eef_pose, goal_pose)
        if error_pose < err_limit:
            return True
        return False


    def _check_support(self, obj_pose):
        obj_mesh = deepcopy(self.obj_mesh_for_sup)
        obj_mesh.apply_transform(obj_pose)
        self.obj_center_point = obj_mesh.center_mass
        locations, _, _ = self.obj_mesh_on_sup.ray.intersects_location(
                    ray_origins=[self.obj_center_point],
                    ray_directions=[[0, 0, -1]])

        if len(locations) != 0:
            support_index = np.where(locations == np.max(locations, axis=0)[2])
            self.obj_support_point = locations[support_index[0]]
            return True
        logger.warning("Not found support point")
        return False


class NotFoundError(Exception):
    """
    Class of custom Exception about Not Found

    Args:
        data (all types): input data
    """
    def __init__(self, data):
        self.data = data
    
    def __str__(self):
        return f"Not Found {self.data}, please check the name again"


class CollisionError(Exception):
    """
    Class of custom Exception about Collision

    Args:
        data (all types): input data
    """
    def __init__(self, data):
        self.data = data
    
    def __str__(self):
        return f"Check the collision.. {self.data}, please check settings again"

class LimitJointError(Exception):
    """
    Class of custom Exception about Collision

    Args:
        data (all types): input data
    """
    def __init__(self, *data):
        self.data = data
    
    def __str__(self):
        return f"Check the joints.. {self.data}, please check current joints setting again"

class OriValueError(Exception):
    """
    Class of custom Exception about Orientation Value

    Args:
        data (all types): input data
    """
    def __init__(self, data):
        self.data = data
    
    def __str__(self):
        return "Expecting the shape of the orientation to be (3,), (3,3), or (4,), instead got:""{}".format(self.data)




import numpy as np
import collections 
import itertools
import copy

try:
    import fcl
    import trimesh
except BaseException:
    fcl = None
    trimesh = None

from pykin.utils.error_utils import CollisionError
from pykin.utils.transform_utils import get_h_mat
from pykin.utils.log_utils import create_logger
from pykin.collision.contact_data import ContactData
logger = create_logger('Collision Manager', "debug",)

class CollisionManager:
    """
    A rigid body collision manager.
    """

    def __init__(self, mesh_path=None):
        self.mesh_path = mesh_path
        self._objs = {}
        self._names = collections.defaultdict(lambda: None)
        self._manager = fcl.DynamicAABBTreeCollisionManager()
        self._manager.setup()
        self._filter_names = set()

    def __repr__(self):
        return 'pykin.collision.collision_manager.{}()'.format(type(self).__name__)

    def setup_robot_collision(self, robot, fk=None, geom="visual"):
        if fk is None:
            fk = robot.init_transformations
        self._filter_contact_names(robot, fk, geom)
        
    def _filter_contact_names(self, robot, fk, geom):            
        is_collision = False

        def _get_collision_datas():
            nonlocal is_collision
            result = []
            for (name1, name2) in self._filter_names:
                index_name1 = list(self._objs.keys()).index(name1)
                index_name2 = list(self._objs.keys()).index(name2)

                if abs(index_name1-index_name2) > 1:
                    for joint in robot.joints.values():
                        if name1 == joint.parent:
                            if joint.dtype == "revolute":
                                print(name1, name2)
                                print(index_name1, index_name2)
                                is_collision = True
                                result.append((name1, name2))
            return result

        for link, transformation in fk.items():
            if geom == "visual":
                if robot.links[link].visual.gtype == "mesh":
                    mesh_name = robot.links[link].visual.gparam.get('filename')
                    file_name = self.mesh_path + mesh_name
                    mesh = trimesh.load_mesh(file_name)
                    A2B = np.dot(transformation.h_mat, robot.links[link].visual.offset.h_mat)
                    self.add_object(robot.links[link].name, "mesh", mesh, A2B)

            if geom == "collision":
                if robot.links[link].collision.gtype == "mesh":
                    mesh_name = robot.links[link].collision.gparam.get('filename')
                    file_name = self.mesh_path + mesh_name
                    mesh = trimesh.load_mesh(file_name)
                    A2B = np.dot(transformation.h_mat, robot.links[link].collision.offset.h_mat)
                    self.add_object(robot.links[link].name, "mesh", mesh, A2B)

        _, names = self.in_collision_internal(return_names=True)
        self._filter_names = copy.deepcopy(names)

        if robot.robot_name == "ur5e":
            return

        collision_datas = _get_collision_datas()
        if is_collision:
            for name1, name2 in collision_datas:
                logger.error(f"{name1} and {name2} is Collision..")
            raise CollisionError(f"Conflict confirmed. Check the joint settings again") 
    
    def add_object(self, 
                   name, 
                   gtype=None,
                   gparam=None,
                   transform=None):
        """
        Add an object to the collision manager.
        If an object with the given name is already in the manager, replace it.

        Args:
            name (str): An identifier for the object
            gtype (str): object type (cylinder, sphere, box)
            gparam (float or tuple): object parameter (radius, length, size)
            transform (np.array): Homogeneous transform matrix for the object
        """
        if gtype is None:
            return

        if transform is None:
            transform = np.eye(4)
        transform = np.asanyarray(transform, dtype=np.float32)
        if transform.shape != (4, 4):
            if transform.shape == (3,):
                transform = get_h_mat(position=transform)
            else:
                raise ValueError('transform must be (4,4)!')

        if gtype == "mesh":
            geom = self._get_BVH(gparam)
        else:
            geom = self._get_geom(gtype, gparam)
        
        t = fcl.Transform(transform[:3, :3], transform[:3, 3])
        o = fcl.CollisionObject(geom, t)

        # Add collision object to set
        if name in self._objs:
            self._manager.unregisterObject(self._objs[name])
        self._objs[name] = {'obj': o,
                            'geom': geom}
        # store the name of the geometry
        self._names[id(geom)] = name

        self._manager.registerObject(o)
        self._manager.update()

    def set_transform(self, name=None, transform=np.eye(4)):
        """
        Set the transform for one of the manager's objects.
        This replaces the prior transform.
        
        Args:
            name (str): An identifier for the object already in the manager
            transform (np.array): A new homogeneous transform matrix for the object
        """
        if name is None:
            return
            
        if name in self._objs:
            o = self._objs[name]['obj']
            o.setRotation(transform[:3, :3])
            o.setTranslation(transform[:3, 3])
            self._manager.update(o)
        else:
            raise ValueError('{} not in collision manager!'.format(name))

    def remove_object(self, name):
        """
        Delete an object from the collision manager.
        
        Args:
            name (str): The identifier for the object
        """
        if name in self._objs:
            self._manager.unregisterObject(self._objs[name]['obj'])
            self._manager.update(self._objs[name]['obj'])
            # remove objects from _objs
            geom_id = id(self._objs.pop(name)['geom'])
            # remove names
            self._names.pop(geom_id)
            print(f"{name} object is removed")
        else:
            raise ValueError('{} not in collision manager!'.format(name))

    def reset_all_object(self):
        """
        Reset all object from the collision manager.
        """
        self._objs = {}
        self._names = collections.defaultdict(lambda: None)
        self._manager = fcl.DynamicAABBTreeCollisionManager()
        self._manager.setup()

    def in_collision_internal(self, return_names=False, return_data=False):
        cdata = fcl.CollisionData()
        if return_names or return_data:
            cdata = fcl.CollisionData(request=fcl.CollisionRequest(
                num_max_contacts=100000, enable_contact=False))

        self._manager.collide(cdata, fcl.defaultCollisionCallback)

        result = cdata.result.is_collision

        objs_in_collision = set()
        contact_data = []

        if return_names or return_data:
            for contact in cdata.result.contacts:
                coll_names = (self._extract_name(contact.o1),self._extract_name(contact.o2))
                coll_names = tuple(sorted(coll_names))

                if (coll_names[0], coll_names[1]) in self._filter_names:
                    continue

                if ("obstacle" in coll_names[0] and "obstacle" in coll_names[1]):
                    continue

                if return_names:
                    objs_in_collision.add(coll_names)
                if return_data:
                    contact_data.append(ContactData(coll_names, contact))

        if not objs_in_collision:
            result = False
            objs_in_collision = "No object collided.."

        return self._get_returns(return_names, return_data, result, objs_in_collision, contact_data)

    def in_collision_other(self, other_manager=None, return_names=False, return_data=False):
        if other_manager is None:
            return 
            
        cdata = fcl.CollisionData()
        if return_names or return_data:
            cdata = fcl.CollisionData(request=fcl.CollisionRequest(
                num_max_contacts=100000, enable_contact=False))

        self._manager.collide(other_manager._manager, cdata, fcl.defaultCollisionCallback)

        result = cdata.result.is_collision

        objs_in_collision = set()
        contact_data = []

        if return_names or return_data:
            for contact in cdata.result.contacts:
                reverse = False
                coll_names = (self._extract_name(contact.o1), other_manager._extract_name(contact.o2))

                if (coll_names[0], coll_names[1]) in self._filter_names:
                    continue

                if ("obstacle" in coll_names[0] and "obstacle" in coll_names[1]):
                    continue

                if coll_names[0] is None:
                    coll_names = (self._extract_name(contact.o2), other_manager._extract_name(contact.o1))
                    reverse = True

                if return_names:
                    objs_in_collision.add(coll_names)
                if return_data:
                    if reverse:
                        coll_names = reversed(coll_names)
                    contact_data.append(ContactData(coll_names, contact))

            if not objs_in_collision:
                result = False
                objs_in_collision = "No object collided.."

        return self._get_returns(return_names, return_data, result, objs_in_collision, contact_data)

    def get_distances_internal(self):
        req = fcl.DistanceRequest()
        res = fcl.DistanceResult()

        result = collections.defaultdict(float)
        for (o1, o2) in list(itertools.combinations(self._objs, 2)):
            if (o1, o2) in self._filter_names:
                continue
            distance = np.round(fcl.distance(self._objs[o1]['obj'],self._objs[o2]['obj'], req, res), 6)
            result[(o1, o2)] = distance
        
        return result

    def get_distances_other(self, other_manager):
        def _mix_objects():
            for o1 in self._objs:
                for o2 in other_manager._objs:
                    yield (o1, o2)

        req = fcl.DistanceRequest()
        res = fcl.DistanceResult()
        
        result = collections.defaultdict(float)
        for (o1, o2) in _mix_objects():
            distance = np.round(fcl.distance(self._objs[o1]['obj'], other_manager._objs[o2]['obj'], req, res), 6)
            result[(o1, o2)] = distance
        
        return result

    @staticmethod
    def _get_returns(return_names, return_data, *args):
        if return_names and return_data:
            return args[0], args[1], args[2]
        elif return_names:
            return args[0], args[1]
        elif return_data:
            return args[0], args[2]
        else:
            return args[0]

    def _get_BVH(self, mesh):
        """
        Get a BVH for a mesh.

        Parameters
        -------------
        mesh : Trimesh
          Mesh to create BVH for

        Returns
        --------------
        bvh : fcl.BVHModel
          BVH object of source mesh
        """
        bvh = self.mesh_to_BVH(mesh)
        return bvh

    @staticmethod
    def mesh_to_BVH(mesh):
        """
        Create a BVHModel object from a Trimesh object

        Parameters
        -----------
        mesh : Trimesh
        Input geometry

        Returns
        ------------
        bvh : fcl.BVHModel
        BVH of input geometry
        """
        bvh = fcl.BVHModel()
        bvh.beginModel(num_tris_=len(mesh.faces),
                    num_vertices_=len(mesh.vertices))
        bvh.addSubModel(verts=mesh.vertices,
                        triangles=mesh.faces)
        bvh.endModel()
        return bvh

    @staticmethod
    def _get_geom(gtype, gparam):
        """
        Get fcl geometry from robot's geometry type or params
        
        Args:
            geom (CollisionObject): Input model
        
        Returns:
            names (hashable): Name of input geometry
        """
        geom = None
        if gtype == "cylinder":
            radius = gparam[0]
            length = gparam[1]
            geom = fcl.Cylinder(radius, length)
        elif gtype == "sphere":
            radius = float(gparam)
            geom = fcl.Sphere(radius)
        elif gtype == "box":
            size = gparam
            geom = fcl.Box(*size)
        elif gtype == "mesh":
            pass
        return geom

    def _extract_name(self, geom):
        """
        Retrieve the name of an object from the manager by its
        CollisionObject, or return None if not found.
        
        Args:
            geom (CollisionObject): Input model
        
        Returns:
            names (hashable): Name of input geometry
        """
        return self._names[id(geom)]





import numpy as np


import pykin.utils.transform_utils as t_utils
import pykin.utils.kin_utils as k_utils
import pykin.kinematics.jacobian as jac

from pykin.planners.planner import Planner
from pykin.utils.error_utils import OriValueError, CollisionError
from pykin.utils.kin_utils import ShellColors as sc, logging_time
from pykin.utils.log_utils import create_logger
from pykin.utils.transform_utils import get_linear_interpoation, get_quaternion_slerp

logger = create_logger('Cartesian Planner', "debug",)

class CartesianPlanner(Planner):
    """
    path planner in Cartesian space

    Args:
        robot(SingleArm or Bimanual): The manipulator robot type is SingleArm or Bimanual
        self_collision_manager: CollisionManager for robot's self collision check
        obstacle_collision_manager: CollisionManager for collision check between robot and object
        n_step(int): Number of waypoints
        dimension(int): robot arm's dof
        waypoint_type(str): Type of waypoint ex) "Linear", "Cubic", "Circular"
    """
    def __init__(
        self,
        robot,
        self_collision_manager=None,
        obstacle_collision_manager=None,
        n_step=500,
        dimension=7,
        waypoint_type="Linear"
    ):
        super(CartesianPlanner, self).__init__(
            robot, 
            self_collision_manager, 
            obstacle_collision_manager,
            dimension)
        self.n_step = n_step
        self.waypoint_type = waypoint_type
        self.eef_name = self.robot.eef_name
        self.arm = None
        self._dimension = dimension

        super()._setup_q_limits()
        super()._setup_eef_name()
    
    def __repr__(self):
        return 'pykin.planners.cartesian_planner.{}()'.format(type(self).__name__)
    
    @logging_time
    def get_path_in_joinst_space(
        self, 
        current_q=None,
        goal_pose=None,
        waypoints=None,
        resolution=1, 
        damping=0.5,
        epsilon=1e-12,
        pos_sensitivity=0.03,
        is_slerp=False
    ):
        self._cur_qpos = super()._change_types(current_q)
        self._goal_pose = super()._change_types(goal_pose)
        init_fk = self.robot.kin.forward_kinematics(self.robot.desired_frames, self._cur_qpos)
        self._cur_pose = self.robot.get_eef_pose(init_fk)

        self._resolution = resolution
        self._damping = damping
        self._pos_sensitivity = pos_sensitivity
        self._is_slerp = is_slerp

        if waypoints is None:
            waypoints = self.generate_waypoints(is_slerp)

        paths, target_positions = self._compute_path_and_target_pose(waypoints, epsilon)
        
        return paths, target_positions

    def _compute_path_and_target_pose(self, waypoints, epsilon):
        cnt = 0
        total_cnt = 10
        while True:
            cnt += 1
            collision_pose = {}
            cur_fk = self.robot.kin.forward_kinematics(self.robot.desired_frames, self._cur_qpos)

            current_transform = cur_fk[self.eef_name].h_mat
            eef_position = cur_fk[self.eef_name].pos

            paths = [self._cur_qpos]
            target_positions = [eef_position]

            for step, (pos, ori) in enumerate(waypoints):
                target_transform = t_utils.get_h_mat(pos, ori)
                err_pose = k_utils.calc_pose_error(target_transform, current_transform, epsilon) 
                J = jac.calc_jacobian(self.robot.desired_frames, cur_fk, self._dimension)
                J_dls = np.dot(J.T, np.linalg.inv(np.dot(J, J.T) + self._damping**2 * np.identity(6)))

                dq = np.dot(J_dls, err_pose)
                self._cur_qpos = np.array([(self._cur_qpos[i] + dq[i]) for i in range(self._dimension)]).reshape(self._dimension,)

                is_collision_free = self._collision_free(self._cur_qpos, visible_name=False)

                if not is_collision_free:
                    collision_pose[step] = np.round(target_transform[:3,3], 6)
                    continue

                if not self._check_q_in_limits(self._cur_qpos):
                    continue

                cur_fk = self.robot.kin.forward_kinematics(self.robot.desired_frames, self._cur_qpos)
                current_transform = cur_fk[self.robot.eef_name].h_mat

                if step % (1/self._resolution) == 0 or step == len(waypoints)-1:
                    paths.append(self._cur_qpos)
                    target_positions.append(pos)

            err = t_utils.compute_pose_error(self._goal_pose[:3], cur_fk[self.eef_name].pos)
            
            if collision_pose.keys():
                logger.error(f"Failed Generate Path.. Collision may occur.")
                
                for pose in collision_pose.values():
                    logger.warning(f"Collision Position : {pose}")
                raise CollisionError("Conflict confirmed. Check the object position!")
                
            if err < self._pos_sensitivity:
                logger.info(f"Generate Path Successfully!! Error is {err:6f}")
                break

            if cnt > total_cnt:
                logger.error(f"Failed Generate Path.. The number of retries of {cnt} exceeded")
                paths, target_positions = None, None
                break

            logger.error(f"Failed Generate Path.. Position Error is {err:6f}")
            print(f"{sc.BOLD}Retry Generate Path, the number of retries is {cnt}/{total_cnt} {sc.ENDC}\n")
            
        return paths, target_positions

    # TODO
    # generate cubic, circular waypoints
    def generate_waypoints(self, is_slerp):
        if self.waypoint_type == "Linear":
            waypoints = [path for path in self._get_linear_path(self._cur_pose, self._goal_pose, is_slerp)]
        if self.waypoint_type == "Cubic":
            pass
        if self.waypoint_type == "Circular":
            pass
        return waypoints

    def get_waypoints(self):
        return self.waypoints

    def _change_pose_type(self, pose):
        ret = np.zeros(7)
        ret[:3] = pose[:3]
        
        if isinstance(pose, (list, tuple)):
            pose = np.asarray(pose)
        ori = pose[3:]

        if ori.shape == (3,):
            ori = t_utils.get_quaternion_from_rpy(ori)
            ret[3:] = ori
        elif ori.shape == (4,):
            ret[3:] = ori
        else:
            raise OriValueError(ori.shape)

        return ret

    def _get_linear_path(self, init_pose, goal_pose, is_slerp):
        for step in range(1, self.n_step + 1):
            delta_t = step / self.n_step
            pos = get_linear_interpoation(init_pose[:3], goal_pose[:3], delta_t)
            ori = init_pose[3:]
            if is_slerp:
                ori = get_quaternion_slerp(init_pose[3:], goal_pose[3:], delta_t)

            yield (pos, ori)

    def _get_cubic_path(self):
        pass

    def _get_cicular_path(self):
        pass

    @property
    def resolution(self):
        return self._resolution
    
    @resolution.setter
    def resolution(self, resolution):
        self._resolution = resolution

    @property
    def damping(self):
        return self._damping
    
    @damping.setter
    def damping(self, damping):
        self._damping = damping

    @property
    def pos_sensitivity(self):
        return self._pos_sensitivity
    
    @pos_sensitivity.setter
    def pos_sensitivity(self, pos_sensitivity):
        self._pos_sensitivity = pos_sensitivity

    @property
    def is_slerp(self):
        return self._is_slerp
    
    @is_slerp.setter
    def is_slerp(self, is_slerp):
        self._is_slerp = is_slerp


import numpy as np
from pykin.utils import transform_utils as tf

class Transform:
    """
    This class calculates the rotation and translation of a 3D rigid body.

    Args:
        pos (sequence of float) : The translation parameter.
        rot (sequence of float) : The rotation parameter. Give in quaternions or roll pitch yaw.
    """
    def __init__(
        self, 
        pos=np.zeros(3),
        rot=np.array([1.0, 0.0, 0.0, 0.0]) 
    ):
        # Set rotation, position
        self.pos = self._to_pos(pos)
        self.rot = self._to_quaternion(rot)

    def __str__(self):
        return "Transform(pos={0}, rot={1})".format(self.pos, self.rot)

    def __repr__(self):
        return 'pykin.kinematics.transform.{}()'.format(type(self).__name__)

    def __mul__(self, other):
        rot = tf.quaternion_multiply(self.rot, other.rot)
        pos = self._to_rotation_vec(self.rot, other.pos) + self.pos
        return Transform(pos, rot)

    def inverse(self):
        """
        Returns:
            Transform : inverse transform
        """
        rot = tf.get_quaternion_inverse(self.rot)
        pos = -self._to_rotation_vec(rot, self.pos)
        return Transform(pos, rot)

    @property
    def pos(self):
        """
        Returns:
            np.array: position
        """
        return self._pos

    @pos.setter
    def pos(self, pos):
        self._pos = self._to_pos(pos)
        
    @property
    def rot(self):
        """
        Returns:
            np.array: rotation (quaternion)
        """
        return self._rot
    
    @rot.setter
    def rot(self, rot):
        self._rot = self._to_quaternion(rot)

    @property
    def pose(self):
        """
        Returns:
            np.array: pose
        """
        return np.hstack((self.pos, self.rot))

    @property
    def rotation_matrix(self):
        """
        Returns:
            np.array: rotation matrix
        """
        return tf.get_rotation_matrix(self.rot)

    @property
    def h_mat(self):
        """
        Returns:
            np.array: homogeneous matrix
        """
        mat = tf.get_h_mat_from_quaternion(self.rot)
        mat[:3, 3] = self.pos
        return mat

    @staticmethod
    def _to_rotation_vec(rot, vec):
        """
        Convert with quaternion and position to rotation vector

        Args:
            rot (np.array): rotation (quaternion)
            vec (np.array): position

        Returns:
            np.array: rotation vector
        """
        v4 = np.hstack([np.array([0.0]), vec])
        inv_rot = tf.get_quaternion_inverse(rot)
        ans = tf.quaternion_multiply(tf.quaternion_multiply(rot, v4), inv_rot)
        return ans[1:]

    @staticmethod
    def _to_quaternion(rot):
        """
        Convert to rotation (qauternion)

        Args:
            rot (sequence of float): rotation (quaternion)

        Returns:
            np.array: rotation (quaternion)
        """
        if len(rot) == 3:
            rot = tf.get_quaternion_from_rpy(rot, convention='wxyz')
        elif len(rot) == 4:
            rot = np.array(rot)
        else:
            raise ValueError("Size of rot must be 3 or 4.")
        return rot

    @staticmethod
    def _to_pos(pos):
        """
        Convert to pos vector

        Args:
            pos (sequence of float): position

        Returns:
            np.array: position
        """
        if not isinstance(pos, np.ndarray):
            if len(pos) == 3:
                pos = np.array(pos)
            else:
                raise ValueError("Size of pos must be 3.")
        assert pos.shape == (3,)
        
        return pos


import math
import numpy as np

from pykin.planners.planner import Planner
from pykin.planners.tree import Tree
from pykin.utils.log_utils import create_logger
from pykin.utils.kin_utils import ShellColors as sc, logging_time
from pykin.utils.transform_utils import get_linear_interpoation

logger = create_logger('RRT Star Planner', "debug")

class RRTStarPlanner(Planner):
    """
    RRT star path planner

    Args:
        robot(SingleArm or Bimanual): The manipulator robot type is SingleArm or Bimanual
        self_collision_manager: CollisionManager for robot's self collision check
        obstacle_collision_manager: CollisionManager for collision check between robot and object
        delta_distance(float): distance between nearest vertex and new vertex
        epsilon(float): 1-epsilon is probability of random sampling
        gamma_RRT_star(int): factor used for search radius
        max_iter(int): maximum number of iterations
        dimension(int): robot arm's dof
        n_step(int): number for n equal divisions between waypoints
    """
    def __init__(
        self, 
        robot,
        self_collision_manager=None,
        obstacle_collision_manager=None,
        delta_distance=0.5,
        epsilon=0.2,
        max_iter=3000,
        gamma_RRT_star=300, # At least gamma_RRT > delta_distance,
        dimension=7,
        n_step=10
    ):
        super(RRTStarPlanner, self).__init__(
            robot, self_collision_manager, 
            obstacle_collision_manager,
            dimension
        )
        self.delta_dis = delta_distance
        self.epsilon = epsilon
        self._max_iter = max_iter
        self.gamma_RRTs = gamma_RRT_star
        
        self._cur_qpos = None
        self._goal_pose = None
        self.T = None
        self.cost = None

        self.arm = None
        self.dimension = dimension
        self.n_step = n_step
        self.eef_name = self.robot.eef_name

        super()._setup_q_limits()
        super()._setup_eef_name()

    def __repr__(self):
        return 'pykin.planners.rrt_star_planner.{}()'.format(type(self).__name__)
    

    @logging_time
    def get_path_in_joinst_space(self, cur_q, goal_pose, max_iter=None, resolution=1):
        """
        Get path in joint space

        Returns:
            path(list) : result path (from start joints to goal joints)
        """
        self._cur_qpos = super()._change_types(cur_q)
        self._goal_pose = super()._change_types(goal_pose)
        
        if max_iter is not None:
            self._max_iter = max_iter

        cnt = 0
        total_cnt = 10

        while True:
            cnt += 1
            for _ in range(total_cnt):
                self.goal_q = self.robot.inverse_kin(np.random.randn(self._dimension), self._goal_pose)
                if self._check_q_in_limits(self.goal_q):
                    break
                if cnt > total_cnt:
                    logger.error(f"Failed Generate Path.. The number of retries of {cnt} exceeded")
                    break
                print(f"{sc.WARNING}Retry compute IK{sc.ENDC}")

            q_paths = None
            self.T = Tree()
            self.cost = {}

            self.T.add_vertex(self._cur_qpos)
            self.cost[0] = 0

            for step in range(self._max_iter):
                if step % 300 == 0 and step !=0:
                    logger.info(f"iter : {step}")
                    
                rand_q = self.random_state()
                if not self._collision_free(rand_q):
                    continue

                nearest_q, nearest_idx = self.nearest_neighbor(rand_q, self.T)
                new_q = self.new_state(nearest_q, rand_q)
    
                if self._collision_free(new_q) and self._check_q_in_limits(new_q):
                    neighbor_indexes = self.get_near_neighbor_indices(new_q)
                    min_cost = self.get_new_cost(nearest_idx, nearest_q, new_q)
                    min_cost, nearest_idx = self.get_minimum_cost_and_index(neighbor_indexes, new_q, min_cost, nearest_idx)
    
                    self.T.add_vertex(new_q)
                    new_idx = len(self.T.vertices) - 1
                    self.cost[new_idx] = min_cost
                    self.T.add_edge([nearest_idx, new_idx])

                    self.rewire(neighbor_indexes, new_q, new_idx)

                    if self.reach_to_goal(new_q):        
                        q_paths = self.find_path(self.T)

            if q_paths is not None:
                logger.info(f"Generate Path Successfully!!")  
                break 

            if cnt > total_cnt:
                logger.error(f"Failed Generate Path.. The number of retries of {cnt} exceeded")
                break

            logger.error(f"Failed Generate Path..")
            print(f"{sc.BOLD}Retry Generate Path, the number of retries is {cnt}/{total_cnt} {sc.ENDC}\n")
        
        result_q_paths = []
        if q_paths is not None:
            for step, joint in enumerate(q_paths):
                if step % round(1/resolution) == 0 or step == len(q_paths)-1:
                    result_q_paths.append(joint)

            interpolate_path = []
            interpolate_paths = []
            
            for i in range(len(result_q_paths)-1):
                interpolate_path = [path.tolist() for path in self._get_linear_path(result_q_paths[i], result_q_paths[i+1])]
                interpolate_paths.extend(interpolate_path)

        return interpolate_paths, result_q_paths

    def random_state(self):
        """
        sampling joints in q space within joint limits 
        If random probability is greater than the epsilon, return random joint angles
        oterwise, return goal joint angles

        Returns:
            q_outs(np.array) : 
        """
        q_outs = np.zeros(self.dimension)
        
        if np.random.random() > self.epsilon:
            for i, (q_min, q_max) in enumerate(zip(self.q_limits_lower, self.q_limits_upper)):
                q_outs[i] = np.random.uniform(q_min, q_max)
        else:
            q_outs = self.goal_q

        return q_outs

    def nearest_neighbor(self, random_q, tree):
        """
        Find nearest neighbor vertex and index from random_q

        Args:
            random_q(np.array): sampled random joint angles 
            tree(Tree): Trees obtained so far

        Returns:
            nearest_vertex(np.array): nearest vertex(joint angles)
            nearest_idx(int): nearest index
        """
        distances = [self.distance(random_q, vertex) for vertex in tree.vertices]
        nearest_idx = np.argmin(distances)
        nearest_vertex = tree.vertices[nearest_idx]
        return nearest_vertex, nearest_idx

    def distance(self, pointA, pointB):
        """
        Get distance from pointA to pointB

        Args:
            pointA(np.array)
            pointB(np.array)
            
        Returns:
            Norm(float or ndarray)
        """

        theta1 = 0
        theta2 = 0
        dx, dy, dist = 0, 0, 0

        for i in range(self.dimension):
            theta1 += pointA[i]
            theta2 += pointB[i]
            dx += np.cos(theta1) - np.cos(theta2)
            dy += np.sin(theta1) - np.sin(theta2)
            dist += np.sqrt(dx * dx + dy * dy)

        return dist * 1 / self.dimension
        # return np.linalg.norm(pointB - pointA)

    def new_state(self, nearest_q, random_q):
        """
        Get new point between nearest vertex and random vertex

        Args:
            nearest_q(np.array): nearest joint angles 
            random_q(np.array): sampled random joint angles 

        Returns:
            new_q(np.array): new joint angles
        """
        if np.equal(nearest_q, random_q).all():
            return nearest_q

        vector = random_q - nearest_q
        dist = self.distance(random_q, nearest_q)
        step = min(self.delta_dis, dist)
        unit_vector = vector / dist
        new_q = nearest_q + unit_vector * step

        return new_q

    def get_near_neighbor_indices(self, q):
        """
        Returns all neighbor indices within the search radius from the new vertex

        Args:
            q(np.array): new joint angles 

        Returns:
            near_indexes(list): all neighbor indices
        """
        card_V = len(self.T.vertices) + 1
        r = self.gamma_RRTs * ((math.log(card_V) / card_V) ** (1/self.dimension))
        search_radius = min(r, self.delta_dis)
        dist_list = [self.distance(vertex, q) for vertex in self.T.vertices]
                                                   
        near_indexes = []
        for idx, dist in enumerate(dist_list):
            if dist <= search_radius:
                near_indexes.append(idx)

        return near_indexes

    def get_new_cost(self, idx, pointA, pointB):
        """
        Returns new cost 

        Args:
            idx(int): neighbor vertex's index
            A(np.array)
            B(np.array)

        Returns:
            cost(float)
        """
        cost = self.cost[idx] + self.distance(pointA, pointB)
        return cost

    def get_minimum_cost_and_index(self, neighbor_indexes, new_q, min_cost, nearest_idx):
        """
        Returns minimum cost and neer vertex index 
        between neighbor vertices and new vertex in search radius

        Args:
            neighbor_indexes: neighbor vertex's index
            new_q(int): new joint angles
            min_cost(np.array): minimum cost
            nearest_idx(np.array): nearest index

        Returns:
            min_cost(float)
            nearest_idx(int)
        """
        for i in neighbor_indexes:
            new_cost = self.get_new_cost(i, new_q, self.T.vertices[i])

            if new_cost < min_cost:
                min_cost = new_cost
                nearest_idx = i

        return min_cost, nearest_idx

    def rewire(self, neighbor_indexes, new_q, new_idx):
        """
        Rewire a new vertex with a neighbor vertex with minimum cost

        Args:
            neighbor_indexes: neighbor vertex's index
            new_q(int): new joint angles
            new_idx(np.array): new joint angles's index
        """
        for i in neighbor_indexes:
            new_cost = self.get_new_cost(new_idx, new_q, self.T.vertices[i])

            if new_cost < self.cost[i]:
                self.cost[i] = new_cost
                self.T.edges[i-1][0] = new_idx

    def reach_to_goal(self, point):
        """
        Check reach to goal
        If reach to goal, return True
        Args:
            point(np.array): joint angles

        Returns:
            bool
        """
        dist = self.distance(point, self.goal_q)
        if dist <= 0.5:
            return True
        return False

    def find_path(self, tree):
        """
        find path result from start index to goal index

        Args:
            tree(Tree): Trees obtained so far

        Returns:
            path(list)
        """
        path = [self.goal_q]
        goal_idx = tree.edges[-1][1]
 
        while goal_idx != 0:
            if not np.allclose(path[0], tree.vertices[goal_idx]):
                path.append(tree.vertices[goal_idx])
            parent_idx = tree.edges[goal_idx-1][0]
            goal_idx = parent_idx
        path.append(self._cur_qpos)

        return path[::-1]

    def get_rrt_tree(self):
        """
        Return obtained RRT Trees

        Returns:
            verteices(list)
        """
        vertices = []
        for edge in self.T.edges:
            from_node = self.T.vertices[edge[0]]
            goal_node = self.T.vertices[edge[1]]
            vertices.append((from_node, goal_node))
        return vertices

    def _get_linear_path(self, init_pose, goal_pose):
        for step in range(1, self.n_step + 1):
            delta_t = step / self.n_step
            qpos = get_linear_interpoation(init_pose, goal_pose, delta_t)
            yield qpos

    @property
    def max_iter(self):
        return self._max_iter

    @max_iter.setter
    def max_iter(self, max_iter):
        self._max_iter = max_iter


import numpy as np

def calc_jacobian(frames: list, transformations: dict, jsize: int) -> np.array:
    """
    Args:
        frames (list): frames to compute jacobian
        transformations (dict): transformations from forward kinematics
        thetas (int): size of joint space

    Returns:
        Jacobian (np.array(6, jsize)): return Jacobian
    """
    target_position = list(transformations.values())[-1].pos
    J = np.zeros((6, jsize))
    n = 0
    for frame in frames:
        if frame.joint.dtype == "revolute":
            n += 1
            w = np.dot(transformations[frame.link.name].h_mat[:3, :3], frame.joint.axis)
            v = np.cross(w, target_position - transformations[frame.link.name].pos)
            J[:, n - 1] = np.hstack((v, w))
        elif frame.joint.dtype == "prismatic":
            n += 1
            w = np.zeros(3)
            v = np.dot(transformations[frame.link.name].h_mat[:3, :3], frame.joint.axis)
            J[:, n - 1] = np.hstack((v, w))
    return J



import sys, os
pykin_path = os.path.abspath(os.path.dirname(__file__)+"/../../" )
sys.path.append(pykin_path)


def get_mesh_path(mesh_path, robot_name):
    result_path = pykin_path + "/asset/urdf/" + robot_name +"/"
    result_path = result_path + mesh_path
    return result_path


import numpy as np
import math
from collections import Iterable

def vector_norm(data, axis=None, out=None):
    """
    Returns length, i.e. Euclidean norm, of ndarray along axis.
    """
    data = np.array(data, dtype=np.float64, copy=True)
    if out is None:
        if data.ndim == 1:
            return math.sqrt(np.dot(data, data))
        data *= data
        out = np.atleast_1d(np.sum(data, axis=axis))
        np.sqrt(out, out)
        return out
    data *= data
    np.sum(data, axis=axis, out=out)
    np.sqrt(out, out)
    

def get_rot_mat_from_homogeneous(h_mat):
    """
    Returns rotation matrix from homogeneous matrix
    """
    return h_mat[:-1, :-1]


def get_pos_mat_from_homogeneous(h_mat):
    """
    Returns position matrix from homogeneous matrix
    """
    return h_mat[:-1,-1]


def get_pose_from_homogeneous(h_mat):
    """
    Returns (7,1) pose from homogeneous matrix
    """
    position = get_pos_mat_from_homogeneous(h_mat)
    orientation = get_quaternion_from_matrix(
        get_rot_mat_from_homogeneous(h_mat))
    return np.hstack((position, orientation))


def get_rpy_from_matrix(R):
    """
    Returns roll pitch, yaw from Rotation matrix
    """
    r = np.arctan2(R[2, 1], R[2, 2])
    p = np.arctan2(-R[2, 0], np.sqrt(R[0, 0] ** 2 + R[1, 0] ** 2))
    y = np.arctan2(R[1, 0], R[0, 0])

    return np.asarray([r, p, y])


def get_rpy_from_quaternion(q, convention='wxyz'):
    """
    Returns roll pitch, yaw from quaternion
    """
    if convention == 'xyzw':
        x, y, z, w = q[0], q[1], q[2], q[3]  # (N,)
    elif convention == 'wxyz':
        w, x, y, z = q[0], q[1], q[2], q[3]  # (N,)
    roll = np.arctan2(2 * (w*x + y*z), 1 - 2 * (x**2 + y**2))   # (N,)
    pitch = np.arcsin(2 * (w*y - z*x))                          # (N,)
    yaw = np.arctan2(2 * (w*z + x*y), 1 - 2 * (y**2 + z**2))    # (N,)
    rpy = np.asarray([roll, pitch, yaw]).T  # (N,3)
    return rpy

def get_matrix_from_rpy(rpy):
    """
    Returns rotation matrix from rpy
    """
    cr, cp, cy = [np.cos(i) for i in rpy]
    sr, sp, sy = [np.sin(i) for i in rpy]
    R = np.array([[cy*cp, cy*sp*sr - sy*cr, cy*sp*cr + sy*sr],
                  [sy*cp, sy*sp*sr + cy*cr, sy*sp*cr - cy*sr],
                  [-sp, cp*sr, cp*cr]])
    return R


def get_matrix_from_axis_angle(axis, angle):
    """
    Returns rotation matrix from axis angle
    """
    x, y, z = axis
    theta = angle
    c, s = np.cos(theta), np.sin(theta)
    v = 1 - c
    R = np.array([[x ** 2 * v + c, x * y * v - z * s, x * z * v + y * s],
                  [x * y * v + z * s, y ** 2 * v + c, y * z * v - x * s],
                  [x * z * v - y * s, y * z * v + x * s, z ** 2 * v + c]])
    return R


def get_matrix_from_quaternion(q, convention='wxyz'):
    """
    Returns rotation matrix from quaternion
    """
    if isinstance(q, Iterable):
        if convention == 'xyzw':
            x, y, z, w = q
        elif convention == 'wxyz':
            w, x, y, z = q
    else:
        raise TypeError
    R = np.array([[2 * (w**2 + x**2) - 1, 2 * (x*y - w*z), 2 * (x*z + w*y)],
                  [2 * (x*y + w*z), 2 * (w**2 + y**2) - 1, 2*(y*z - w*x)],
                  [2 * (x*z - w*y), 2 * (y*z + w*x), 2 * (w**2 + z**2) - 1]])
    return R


def get_quaternion_from_rpy(rpy, convention='wxyz'):
    """
    Returns quaternion from rpy
    """
    rpy = np.asarray(rpy)
    multiple_rpy = True
    if len(rpy.shape) < 2:
        multiple_rpy = False
        rpy = np.array([rpy])  # (1,3)

    r, p, y = rpy[:, 0], rpy[:, 1], rpy[:, 2]
    cr, sr = np.cos(r/2.), np.sin(r/2.)
    cp, sp = np.cos(p/2.), np.sin(p/2.)
    cy, sy = np.cos(y/2.), np.sin(y/2.)

    w = cr * cp * cy + sr * sp * sy  # (N,)
    x = sr * cp * cy - cr * sp * sy  # (N,)
    y = cr * sp * cy + sr * cp * sy  # (N,)
    z = cr * cp * sy - sr * sp * cy  # (N,)

    if convention == 'xyzw':
        q = np.vstack([x, y, z, w]).T
    elif convention == 'wxyz':
        q = np.vstack([w, x, y, z]).T
    else:
        raise NotImplementedError(
            "Asking for a convention that has not been implemented")

    if not multiple_rpy:
        return q[0]
    return q


def get_quaternion_from_matrix(R, convention='wxyz'):
    """
    Returns quaternion from rotation matrix
    """
    w = 1./2 * np.sqrt(R[0, 0] + R[1, 1] + R[2, 2] + 1)
    x, y, z = 1./2 * np.array([np.sign(R[2, 1] - R[1, 2]) * np.sqrt(R[0, 0] - R[1, 1] - R[2, 2] + 1),
                               np.sign(R[0, 2] - R[2, 0]) *
                               np.sqrt(R[1, 1] - R[2, 2] - R[0, 0] + 1),
                               np.sign(R[1, 0] - R[0, 1]) * np.sqrt(R[2, 2] - R[0, 0] - R[1, 1] + 1)])

    if convention == 'xyzw':
        return np.array([x, y, z, w])
    elif convention == 'wxyz':
        return np.array([w, x, y, z])
    else:
        raise NotImplementedError(
            "Asking for a convention that has not been implemented")


def get_quaternion_from_axis_angle(axis, angle, convention='wxyz'):
    """
    Returns quaternion from axis angle
    """
    w = np.cos(angle / 2.)
    x, y , z = np.sin(angle / 2.) * axis
    if convention == 'xyzw':
        return np.array([x, y, z, w])
    elif convention == 'wxyz':
        return np.array([w, x, y, z])
    else:
        raise NotImplementedError("Asking for a convention that has not been implemented")


def get_quaternion_inverse(quaternion):
    """
    Returns quaternion inverse
    """
    q = np.array(quaternion, dtype=np.float64, copy=True)
    np.negative(q[1:], q[1:])
    return q / np.dot(q, q)


def get_quaternion_slerp(qA, qB, t):
    """
    referred https://en.wikipedia.org/wiki/Slerp#Quaternion_Slerp
    """
    if isinstance(qA, (np.ndarray, list)) and isinstance(qB, (np.ndarray, list)):
        qA = np.asarray(qA)
        qB = np.asarray(qB)
        if qA.shape != (4,):  # quaternion
            raise ValueError("Expecting the shape of the orientation to be (4,)")
        if qB.shape == (0,):
            qB = np.asarray([1.0, 0.0, 0.0, 0.0])
            
        a = qB * get_quaternion_inverse(qA)
        result = np.sign(a) * (np.abs(a)) ** t * qA
        return result


def get_linear_interpoation(postionA, postionB, step):
    postionA = np.asarray(postionA)
    postionB = np.asarray(postionB)
    return postionB * step + postionA * (1 - step)
    

def get_h_mat_from_quaternion(quaternion):
    """
    Returns homogeneous rotation matrix from quaternion.
    """
    q = np.array(quaternion, dtype=np.float64, copy=True)
    n = np.dot(q, q)
    if n < _EPS:
        return np.identity(4)
    q *= math.sqrt(2.0 / n)
    q = np.outer(q, q)
    return np.array([
        [1.0 - q[2, 2] - q[3, 3], q[1, 2] - q[3, 0], q[1, 3] + q[2, 0], 0.0],
        [q[1, 2] + q[3, 0], 1.0 - q[1, 1] - q[3, 3], q[2, 3] - q[1, 0], 0.0],
        [q[1, 3] - q[2, 0], q[2, 3] + q[1, 0], 1.0 - q[1, 1] - q[2, 2], 0.0],
        [0.0, 0.0, 0.0, 1.0]])


def quaternion_multiply(quaternion1, quaternion0):
    """
    Returns multiplication of two quaternions.
    """
    w0, x0, y0, z0 = quaternion0
    w1, x1, y1, z1 = quaternion1
    return np.array(
        [
            -x1 * x0 - y1 * y0 - z1 * z0 + w1 * w0,
            x1 * w0 + y1 * z0 - z1 * y0 + w1 * x0,
            -x1 * z0 + y1 * w0 + z1 * x0 + w1 * y0,
            x1 * y0 - y1 * x0 + z1 * w0 + w1 * z0,
        ],
        dtype=np.float64,
    )

# epsilon for testing whether a number is close to zero
_EPS = np.finfo(float).eps * 4.0

def get_quaternion_about_axis(angle, axis):
    """
    Returns quaternion for rotation about axis.
    """
    q = np.array([0.0, axis[0], axis[1], axis[2]])
    qlen = vector_norm(q)
    if qlen > _EPS:
        q *= math.sin(angle / 2.0) / qlen
    q[0] = math.cos(angle / 2.0)
    return q


def get_h_mat(position=np.zeros(3), orientation=np.array([1.0, 0.0, 0.0, 0.0])):
    """
    Returns homogeneous matrix from position and orientation
    """
    position = np.asarray(position)
    orientation = np.asarray(orientation)
    if orientation.shape == (3,):  # RPY Euler angles
        R = get_matrix_from_rpy(orientation)
    elif orientation.shape == (4,):  # quaternion in the form [x,y,z,w]
        R = get_matrix_from_quaternion(orientation)
    elif orientation.shape == (3, 3):  # Rotation matrix
        R = orientation

    H = np.vstack((np.hstack((R, position.reshape(-1, 1))),
                  np.array([[0, 0, 0, 1]])))
    return H


def get_inverse_homogeneous(matrix):
    """
    Returns homogeneous inverse
    """
    R = matrix[:3, :3].T
    p = -R.dot(matrix[:3, 3].reshape(-1,1))
    return np.vstack((np.hstack((R,p)),
                      np.array([[0, 0, 0, 1]])))


def get_identity_h_mat():
    """
    Returns identity matrix
    """
    return np.identity(4)


def homogeneous_to_pose(matrix):
    """
    Returns pose from h_mat
    """
    position = matrix[:3, -1]
    quaternion = get_quaternion_from_matrix(matrix[:3, :3])
    return np.concatenate((position, quaternion))


def pose_to_homogeneous(pose):
    """
    Returns h_mat from pose
    """
    pose = np.array(pose).flatten()
    position, orientation = pose[:3], pose[3:]
    return get_h_mat(position=position, orientation=orientation)


def get_quaternion(orientation, convention='wxyz'):
    """
    Returns quaternion from orientation
    """
    if isinstance(orientation, tuple) and len(orientation) == 2:
        angle, axis = orientation
        if isinstance(axis, (float, int)) and isinstance(angle, np.ndarray):
            angle, axis = axis, angle
        return get_quaternion_from_axis_angle(axis, angle, convention)
    if isinstance(orientation, (np.ndarray, list)):
        orientation = np.asarray(orientation)
        if orientation.shape == (3,):
            return get_quaternion_from_rpy(orientation, convention)
        if orientation.shape == (4,):
            if convention == 'wxyz':
                x, y, z, w = orientation
                return np.array([w, x, y, z])
            return orientation
        if orientation.shape == (3, 3):
            return get_quaternion_from_matrix(orientation, convention)
        else:
            raise ValueError("Expecting the shape of the orientation to be (3,), (3,3), or (4,), instead got: "
                             "{}".format(orientation.shape))
    else:
        raise TypeError("Expecting the given orientation to be a np.ndarray, quaternion, tuple or list, instead got: "
                        "{}".format(type(orientation)))


def get_rotation_matrix(orientation):
    """
    Returns rotation matrix from orientation
    """
    if isinstance(orientation, tuple) and len(orientation) == 2:
        angle, axis = orientation
        if isinstance(axis, (float, int)) and isinstance(angle, np.ndarray):
            angle, axis = axis, angle
        return get_matrix_from_axis_angle(axis, angle)
    if isinstance(orientation, (np.ndarray, list)):
        orientation = np.asarray(orientation)
        if orientation.shape == (3,):  # RPY Euler angles
            return get_matrix_from_rpy(orientation)
        if orientation.shape == (4,):  # quaternion
            return get_matrix_from_quaternion(orientation)
        if orientation.shape == (3, 3):  # rotation matrix
            return orientation
        else:
            raise ValueError("Expecting the shape of the orientation to be (3,), (3,3), or (4,), instead got: "
                         "{}".format(orientation.shape))
    else:
        raise TypeError("Expecting the given orientation to be a np.ndarray, quaternion, tuple or list, instead got: "
                    "{}".format(type(orientation)))


def compute_pose_error(target=np.eye(4), result=np.eye(4)):
    """
    Computes pose(homogeneous transform) error 

    Args:
        target_HT (np.array): target homogeneous transform
        result_HT (np.array): result homogeneous transform 

    Returns:
        error (np.array)
    """
    if  target.shape == (3,) and result.shape == (3,):
        error = np.linalg.norm(target - result)
        # target = get_h_mat(position = target)
        # result = get_h_mat(position = result)
        return error
    error = np.linalg.norm(
        np.dot(result, np.linalg.inv(target)) - np.mat(np.eye(4)))
    return error


def get_transform_to_visual(transform, visual_transform):
    return np.dot(transform, visual_transform)


def change_to_pose(pose_list):
    if isinstance(pose_list, list):
        result = np.asarray(pose_list)
        
    result = np.zeros(7)

    if len(pose_list) == 7:
        result = pose_list

    if len(pose_list) == 6:
        result[0] = pose_list[0]
        result[1] = pose_list[1]
        result[2] = pose_list[2]
        q = get_quaternion(pose_list[3:])
        result[3] = q[0]
        result[4] = q[1]
        result[5] = q[2]
        result[6] = q[3]

    return result

        



import numpy as np
np.seterr(divide='ignore', invalid='ignore')
import trimesh

import pykin.utils.transform_utils as t_utils

def normalize(vec):
    return vec / np.linalg.norm(vec)

def surface_sampling(mesh, n_samples=2, face_weight=None):
    vertices, face_ind = trimesh.sample.sample_surface(mesh, count=n_samples, face_weight=face_weight)
    normals = mesh.face_normals[face_ind]
    return vertices, face_ind, normals

def projection(v, u):
    return np.dot(v, u) / np.dot(u, u) * u

def get_transform(A, B):
    # TA = B
    # T = B * inv(A)
    return np.dot(B, np.linalg.inv(A))

def get_rotation_from_vectors(A, B):
    unit_A = A / np.linalg.norm(A)
    unit_B = B / np.linalg.norm(B)
    dot_product = np.dot(unit_A, unit_B)
    angle = np.arccos(dot_product)

    rot_axis = np.cross(unit_B, unit_A)
    R = t_utils.get_matrix_from_axis_angle(rot_axis, angle)

    return R


import numpy as np

from pykin.robots.robot import Robot
from pykin.utils.error_utils import NotFoundError

class SingleArm(Robot):
    """
    Initializes a single-armed robot simulation object.

    Args:
        fname (str): path to the urdf file.
        offset (Transform): robot init offset
    """
    def __init__(
        self,
        fname:str,
        offset=None
    ):
        super(SingleArm, self).__init__(fname, offset)
        self._base_name = ""
        self._eef_name  = ""
        self.desired_base_frame = ""
        self._set_joint_limits_upper_and_lower()

    def _set_joint_limits_upper_and_lower(self):
        """
        Set joint limits upper and lower
        """
        for joint, (limit_lower, limit_upper) in self.joint_limits.items():
            if "head" in joint:
                continue
            if self.joints[joint].dtype == "revolute":
                if limit_lower is None and limit_upper is None:
                    limit_lower = -np.pi
                    limit_upper = np.pi
                self.joint_limits_lower.append(limit_lower)
                self.joint_limits_upper.append(limit_upper)

    def check_limit_joint(self, q_in):
        """
        check q_in within joint limits
        If q_in is in joint limits, return True
        otherwise, return False

        Returns:
            bool(True or False)
        """
        return np.all([q_in >= self.joint_limits_lower, q_in <= self.joint_limits_upper])

    def setup_link_name(self, base_name="", eef_name=None):
        """
        Sets robot's desired frame

        Args:
            base_name (str): reference link name
            eef_name (str): end effector name
        """
        self._check_link_name(base_name, eef_name)
        self._base_name = base_name
        self._eef_name = eef_name
        self._set_desired_base_frame()
        self._set_desired_frame()

    def _check_link_name(self, base_name, eef_name):
        """
        Check link name

        Args:
            base_name (str): reference link name
            eef_name (str): end effector name
        """
        if base_name and not base_name in self.links.keys():
            raise NotFoundError(base_name)

        if eef_name is not None and eef_name not in self.links.keys():
            raise NotFoundError(eef_name)

    def _set_desired_base_frame(self):
        """
        Sets robot's desired base frame

        Args:
            arm (str): robot arm (right or left)
        """
        if self.base_name == "":
            self.desired_base_frame = self.root
        else:
            self.desired_base_frame = self.find_frame(
                self.base_name + "_frame")

    def _set_desired_frame(self):
        """
        Sets robot's desired frame

        Args:
            arm (str): robot arm (right or left)
        """
        self.desired_frames = super().generate_desired_frame_recursive(self.desired_base_frame, self.eef_name)
        self._revolute_joint_names = sorted(self.get_revolute_joint_names(self.desired_frames))

    def inverse_kin(self, current_joints, target_pose, method="LM", maxIter=1000):
        """
        Returns joint angles obtained by computing IK
        
        Args:
            current_joints (sequence of float): input joint angles
            target_pose (np.array): goal pose to achieve
            method (str): two methods to calculate IK (LM: Levenberg-marquardt, NR: Newton-raphson)
            maxIter (int): Maximum number of calculation iterations

        Returns:
            joints (np.array): target joint angles
        """
        joints = self.kin.inverse_kinematics(
            self.desired_frames,
            current_joints,
            target_pose,
            method,
            maxIter)
        return joints

    def get_eef_pose(self, transformations=None):
        """
        Get end effector's pose

        Args:
            transformations(OrderedDict)
        
        Returns:
            vals(dict)
        """
        if transformations is None:
            transformations = self.init_transformations

        return np.concatenate((transformations[self.eef_name].pos, transformations[self.eef_name].rot))

    def get_eef_h_mat(self, transformations=None):
        """
        Get end effector's homogeneous marix

        Args:
            transformations(OrderedDict)
        
        Returns:
            vals(dict)
        """
        if transformations is None:
            transformations = self.init_transformations

        return transformations[self.eef_name].h_mat

    def get_eef_pos(self, transformations=None):
        """
        Get end effector's position

        Args:
            transformations(OrderedDict)
        
        Returns:
            vals(dict)
        """
        if transformations is None:
            transformations = self.init_transformations

        return transformations[self.eef_name].pos

    def get_eef_ori(self, transformations=None):
        """
        Get end effector's orientation

        Args:
            transformations(OrderedDict)
        
        Returns:
            vals(dict)
        """
        if transformations is None:
            transformations = self.init_transformations

        return transformations[self.eef_name].rot

    def get_result_qpos(self, init_qpos, eef_pose, n_trials=5):
        cnt = 0
        is_limit_qpos = False
        while cnt <= n_trials:
            result_qpos = self.inverse_kin(np.random.randn(len(init_qpos)), eef_pose, "LM", 500)
            is_limit_qpos = self.check_limit_joint(result_qpos)
            if is_limit_qpos:
                break
            cnt += 1

        if not is_limit_qpos:
            return None
        return result_qpos

    @property
    def base_name(self):
        return self._base_name
        
    @property
    def eef_name(self):
        return self._eef_name

    @property
    def active_joint_names(self):
        return self._revolute_joint_names

    @property
    def arm_dof(self):
        return len([ joint for joint in self.get_revolute_joint_names() if "head" not in joint])



import logging
import logging.handlers

LOG_LEVEL = { "INFO" : logging.INFO, 
              "DEBUG" : logging.DEBUG, 
              "WARNING" : logging.WARNING, 
              "ERROR" : logging.ERROR, 
              "CRITICAL" :logging.CRITICAL 
            }

class CustomFormatter(logging.Formatter):

    grey = "\x1b[38;21m"
    green='\033[1;32m'
    yellow = "\x1b[33;21m"
    red = "\x1b[31;21m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = '[%(levelname)s] [%(name)s]: %(message)s'
    # format = '[%(levelname)s|%(name)s|%(filename)s:%(lineno)s] %(asctime)s >> %(message)s'

    FORMATS = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: green + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt,"%Y-%m-%d %H:%M:%S")
        return formatter.format(record)

def create_logger(logger_name, logging_level="debug", file_name="test.log", is_save=False):
    # Create Logger

    format = '[%(levelname)s] [%(name)s]: %(message)s'
    # |%(filename)s:%(lineno)s] %(asctime)s 

    if logging_level.upper() in LOG_LEVEL.keys():
        level = LOG_LEVEL.get(logging_level.upper())

    logger = logging.getLogger(logger_name)
 
    # Check handler exists
    if len(logger.handlers) > 0:
        return logger # Logger already exists
 
    logger.setLevel(level)

    formatter = logging.Formatter(format, "%Y-%m-%d %H:%M:%S")

    # Create Handlers
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(level)
    stream_handler.setFormatter(CustomFormatter())
    logger.addHandler(stream_handler)

    if is_save:
        rotating_handler = logging.handlers.TimedRotatingFileHandler(file_name, when='h', interval=1, backupCount=0) 
        rotating_handler.setFormatter(formatter)
        logger.addHandler(rotating_handler) 

    return logger




import trimesh
import numpy as np

def apply_robot_to_collision_manager(c_manager, robot, fk=None, geom="visual"):
    if fk is None:
        fk = robot.init_transformations

    for link, transformation in fk.items():
        if geom == "visual":
            if robot.links[link].visual.gtype == "mesh":
                mesh_name = robot.links[link].visual.gparam.get('filename')
                file_name = c_manager.mesh_path + mesh_name
                mesh = trimesh.load_mesh(file_name)
                A2B = np.dot(transformation.h_mat, robot.links[link].visual.offset.h_mat)
                if c_manager._filter_names:
                    c_manager.set_transform(name=robot.links[link].name, transform=A2B)
                else:
                    c_manager.add_object(
                        name=robot.links[link].name, 
                        gtype="mesh",
                        gparam=mesh,
                        transform=A2B)

        if geom == "collision":
            if robot.links[link].collision.gtype == "mesh":
                mesh_name = robot.links[link].collision.gparam.get('filename')
                file_name = c_manager.mesh_path + mesh_name
                mesh = trimesh.load_mesh(file_name)
                A2B = np.dot(transformation.h_mat, robot.links[link].collision.offset.h_mat)
                if c_manager._filter_names:
                    c_manager.set_transform(name=robot.links[link].name, transform=A2B)
                else:
                    c_manager.add_object(
                        name=robot.links[link].name, 
                        gtype="mesh",
                        gparam=mesh,
                        transform=A2B)
    return c_manager


def apply_robot_to_scene(mesh_path=None, scene=None, robot=None, fk=None, geom="visual"):
    if scene is None:
        scene = trimesh.Scene()

    if fk is None:
        fk = robot.init_transformations

    for link, transformation in fk.items():
        if geom == "visual":
            if robot.links[link].visual.gtype == "mesh":
                mesh_name = robot.links[link].visual.gparam.get('filename')
                file_name = mesh_path + mesh_name
                mesh = trimesh.load_mesh(file_name)
                A2B = np.dot(transformation.h_mat, robot.links[link].visual.offset.h_mat)
                color = robot.links[link].visual.gparam.get('color')

                if color is None:
                    color = np.array([0.2, 0, 0])
                else:
                    color = np.array([color for color in color.values()]).flatten()

                mesh.visual.face_colors = color
                scene.add_geometry(mesh, transform=A2B)  

        if geom == "collision":
            if robot.links[link].collision.gtype == "mesh":
                mesh_name = robot.links[link].collision.gparam.get('filename')
                file_name = mesh_path + mesh_name
                mesh = trimesh.load_mesh(file_name)
                A2B = np.dot(transformation.h_mat, robot.links[link].collision.offset.h_mat)
                color = robot.links[link].collision.gparam.get('color')

                if color is None:
                    color = np.array([0.2, 0, 0])
                else:
                    color = np.array([color for color in color.values()]).flatten()
                
                mesh.visual.face_colors = color
                scene.add_geometry(mesh, transform=A2B)      
    scene.set_camera(np.array([np.pi/2, 0, np.pi/2]), 5, resolution=(1024, 512))
    return scene


def get_mesh_param(link_type):
    file_name = str(link_type.gparam.get('filename'))
    color = link_type.gparam.get('color')
    color = np.array([color for color in color.values()]).flatten()
    return (file_name, color)

def get_cylinder_param(link_type):
    radius = float(link_type.gparam.get('radius'))
    length = float(link_type.gparam.get('length'))
    return (radius, length)

def get_spehre_param(link_type):
    radius = float(link_type.gparam.get('radius'))
    return radius

def get_box_param(link_type):
    size = list(link_type.gparam.get('size'))
    return size

def get_robot_collision_geom(link):
    """
    Get robot collision geometry from link

    Args:
        link (Link): robot's link

    Returns:
        name (str): geom's name
        gtype: geom's type
        gparam: geom's param
    """
    name = None
    gtype = None
    gparam = None

    if link.collision.gtype == "cylinder":
        name = link.name
        gtype = link.collision.gtype
        gparam = get_cylinder_param(link.collision)
    elif link.collision.gtype == "sphere":
        name = link.name
        gtype = link.collision.gtype
        gparam = get_spehre_param(link.collision)
    elif link.collision.gtype == "box":
        name = link.name
        gtype = link.collision.gtype
        gparam = get_box_param(link.collision)

    return name, gtype, gparam

def get_robot_visual_geom(link):
    """
    Get robot visual geometry from link

    Args:
        link (Link): robot's link

    Returns:
        name (str): geom's name
        gtype: geom's type
        gparam: geom's param
    """

    name = None
    gtype = None
    gparam = None

    if link.visual.gtype == "mesh":
        name = link.name
        gtype = link.visual.gtype
        gparam = get_cylinder_param(link.visual)
    if link.visual.gtype == "cylinder":
        name = link.name
        gtype = link.visual.gtype
        gparam = get_cylinder_param(link.visual)
    elif link.visual.gtype == "sphere":
        name = link.name
        gtype = link.visual.gtype
        gparam = get_spehre_param(link.visual)
    elif link.visual.gtype == "box":
        name = link.name
        gtype = link.visual.gtype
        gparam = get_box_param(link.visual)

    return name, gtype, gparam







import numpy as np

class Tree:
    """
    Tree class
    """
    def __init__(self):
        self.vertices = []
        self.edges = []

    def add_vertex(self, q_joints):
        """
        Add q_joints in vertices
        """
        self.vertices.append(q_joints)
        
    def add_edge(self, q_joints_idx):
        """
        Add vertex indexes(parent, child) in edges
        """
        self.edges.append(q_joints_idx)



from collections import OrderedDict

class RobotModel:
    """
    Base class of robot model from urdf file
    When initialized, loads links and joints of robot.
    """
    def __init__(self):
        self._links = OrderedDict()
        self._joints = OrderedDict()

    def find_frame(self):
        """
        Find robot's frame
        """
        raise NotImplementedError

    def find_link(self):
        """
        Find robot's link
        """
        raise NotImplementedError

    def find_joint(self):
        """
        Find robot's joint
        """
        raise NotImplementedError

    @property
    def links(self):
        """
        Returns:
            OrderedDict: all links
        """
        return self._links

    @property
    def joints(self):
        """
        Returns:
            OrderedDict: all joints
        """
        return self._joints

    @property
    def dof(self):
        """
        Robot's dof
        """
        raise NotImplementedError

    @property
    def num_links(self):
        """
        Number of links
        """
        raise NotImplementedError

    @property
    def num_joints(self):
        """
        Number of joints
        """
        raise NotImplementedError

    @property
    def num_fixed_joints(self):
        """
        Number of fixed joints
        """
        raise NotImplementedError

    @property
    def num_actuated_joints(self):
        """
        Number of actuated(revolute or prismatic) joints
        """
        raise NotImplementedError

    @property
    def num_revolute_joints(self):
        """
        Number of revolute joints
        """
        raise NotImplementedError


from pykin.utils.kin_utils import convert_string_to_narray, LINK_TYPES

class URDF_Link:
    """
    Class of parsing link info described in URDF
    """
    @staticmethod
    def set_visual(elem_link, link_frame):
        """
        Set link visual
        """ 
        for elem_visual in elem_link.findall('visual'):
            URDF_Link.set_visual_origin(elem_visual, link_frame)
            URDF_Link.set_visual_geometry(elem_visual, link_frame)
            URDF_Link.set_visual_color(elem_visual, link_frame)

    @staticmethod
    def set_collision(elem_link, link_frame):
        """
        Set link collision
        """ 
        for elem_collision in elem_link.findall('collision'):
            URDF_Link.set_collision_origin(elem_collision, link_frame)
            URDF_Link.set_collision_geometry(elem_collision, link_frame)
            URDF_Link.set_collision_color(elem_collision, link_frame)
            
    @staticmethod
    def set_visual_origin(elem_visual, frame):
        """
        Set link visual's origin
        """ 
        for elem_origin in elem_visual.findall('origin'):
            frame.link.visual.offset.pos = convert_string_to_narray(elem_origin.attrib.get('xyz'))
            frame.link.visual.offset.rot = convert_string_to_narray(elem_origin.attrib.get('rpy'))

    @staticmethod
    def set_visual_geometry(elem_visual, frame):
        """
        Set link visual's geometry
        """ 

        def _set_link_visual_geom(shapes, frame):
            """
            Set link visual's geometry
            """ 
            if shapes.tag == "box":
                frame.link.visual.gtype = shapes.tag
                frame.link.visual.gparam = {"size" : convert_string_to_narray(shapes.attrib.get('size', None))}
            elif shapes.tag == "cylinder":
                frame.link.visual.gtype = shapes.tag
                frame.link.visual.gparam = {"length" : shapes.attrib.get('length', 0),
                                            "radius" : shapes.attrib.get('radius', 0)}
            elif shapes.tag == "sphere":
                frame.link.visual.gtype = shapes.tag
                frame.link.visual.gparam = {"radius" : shapes.attrib.get('radius', 0)}
            elif shapes.tag == "mesh":
                frame.link.visual.gtype = shapes.tag
                frame.link.visual.gparam = {"filename" : shapes.attrib.get('filename', None)}
            else:
                frame.link.visual.gtype = None
                frame.link.visual.gparam = None

        for elem_geometry in elem_visual.findall('geometry'):
            for shape_type in LINK_TYPES:
                for shapes in elem_geometry.findall(shape_type):
                    _set_link_visual_geom(shapes, frame)

    @staticmethod
    def set_visual_color(elem_visual, frame):
        """
        Set link visual's color
        """ 
        for elem_matrial in elem_visual.findall('material'):
            for elem_color in elem_matrial.findall('color'):
                rgba = convert_string_to_narray(elem_color.attrib.get('rgba'))
                frame.link.visual.gparam['color'] = {elem_matrial.get('name') : rgba}
    
    @staticmethod
    def set_collision_origin(elem_collision, frame):
        """
        Set link collision's origin
        """ 
        for elem_origin in elem_collision.findall('origin'):
            frame.link.collision.offset.pos = convert_string_to_narray(elem_origin.attrib.get('xyz'))
            frame.link.collision.offset.rot = convert_string_to_narray(elem_origin.attrib.get('rpy'))

    @staticmethod
    def set_collision_geometry(elem_collision, frame):
        """
        Set link collision's geometry
        """ 
        
        def _set_link_collision_geom(shapes, frame):
            if shapes.tag == "box":
                frame.link.collision.gtype = shapes.tag
                frame.link.collision.gparam = {"size" : convert_string_to_narray(shapes.attrib.get('size', None))}
            elif shapes.tag == "cylinder":
                frame.link.collision.gtype = shapes.tag
                frame.link.collision.gparam = {"length" : shapes.attrib.get('length', 0),
                                            "radius" : shapes.attrib.get('radius', 0)}
            elif shapes.tag == "sphere":
                frame.link.collision.gtype = shapes.tag
                frame.link.collision.gparam = {"radius" : shapes.attrib.get('radius', 0)}
            elif shapes.tag == "mesh":
                frame.link.collision.gtype = shapes.tag
                frame.link.collision.gparam = {"filename" : shapes.attrib.get('filename', None)}
            else:
                frame.link.collision.gtype = None
                frame.link.collision.gparam = None

        elem_geometry = elem_collision.find('geometry')
        for shape_type in LINK_TYPES:
            for shapes in elem_geometry.findall(shape_type):
                _set_link_collision_geom(shapes, frame)

    @staticmethod
    def set_collision_color(elem_collision, frame):
        """
        Set link visual's color
        """ 
        for elem_matrial in elem_collision.findall('material'):
            for elem_color in elem_matrial.findall('color'):
                rgba = convert_string_to_narray(elem_color.attrib.get('rgba'))
                frame.link.collision.gparam['color'] = {elem_matrial.get('name') : rgba}
    


import numpy as np
from pykin.robots.robot import Robot
from pykin.utils.error_utils import NotFoundError

class Bimanual(Robot):
    """
    Initializes a bimanual robot simulation object.

    Args:
        fname (str): path to the urdf file.
        offset (Transform): robot init offset
    """
    def __init__(
        self,
        fname: str,
        offset=None
    ):
        super(Bimanual, self).__init__(fname, offset)
        self._setup_input2dict()
        self._set_joint_limits_upper_and_lower()

    def _setup_input2dict(self):
        """
        Setup dictionary name
        """
        self._base_name = self._input2dict("")
        self._eef_name = {}
        self.desired_base_frame = self._input2dict(None)
        self.desired_frames = self._input2dict(None)
        self._frames = self._input2dict(None)
        self._revolute_joint_names = self._input2dict(None)
        self._target_pose = self._input2dict(None)
        self.joint_limits_lower = self._input2dict(None)
        self.joint_limits_upper = self._input2dict(None)

    def _input2dict(self, inp):
        """
        Helper function that converts an input that is either a single value or a list into a dict with keys for
        each arm: "right", "left"

        Args:
            inp (str or list or None): Input value to be converted to dict
            :Note: If inp is a list, then assumes format is [right, left]

        Returns:
            dict: Inputs mapped for each robot arm
        """
        # First, convert to list if necessary
        if not isinstance(inp, list):
            inp = [inp for _ in range(2)]
        # Now, convert list to dict and return
        return {key: value for key, value in zip(self._arms, inp)}

    def _set_joint_limits_upper_and_lower(self):
        """
        Set joint limits upper and lower
        """
        limits_lower = []
        limits_upper = []

        for joint, (limit_lower, limit_upper) in self.joint_limits.items():
            limits_lower.append((joint, limit_lower))
            limits_upper.append((joint, limit_upper))

        for arm in self._arms:
            self.joint_limits_lower[arm] = [
                limit_lower for joint, limit_lower in limits_lower if arm in joint]
            self.joint_limits_upper[arm] = [
                limit_upper for joint, limit_upper in limits_upper if arm in joint]

    def setup_link_name(self, base_name="", eef_name=None):
        """
        Sets robot's link name

        Args:
            base_name (str): reference link name
            eef_name (str): end effector name
        """
        if "right" in eef_name:
            self._base_name["right"] = base_name
            self._eef_name["right"] = eef_name
            self._set_desired_base_frame("right")
            self._set_desired_frame("right")
            
        if "left" in eef_name:
            self._base_name["left"] = base_name
            self._eef_name["left"] = eef_name
            self._set_desired_base_frame("left")
            self._set_desired_frame("left")

    def _set_desired_base_frame(self, arm):
        """
        Sets robot's desired base frame

        Args:
            arm (str): robot arm (right or left)
        """
        if self.base_name[arm] == "":
            self.desired_base_frame[arm] = self.root
        else:
            self.desired_base_frame[arm] = super().find_frame(self.base_name[arm] + "_frame")

    def _set_desired_frame(self, arm):
        """
        Sets robot's desired frame

        Args:
            arm (str): robot arm (right or left)
        """
        self.desired_frames[arm] = super().generate_desired_frame_recursive(
            self.desired_base_frame[arm],
            self.eef_name[arm])

        self._frames[arm] = self.desired_frames[arm]
        self._revolute_joint_names[arm] = super().get_revolute_joint_names(self._frames[arm])
        self._target_pose[arm] = np.zeros(len(self._revolute_joint_names[arm]))

    def inverse_kin(self, current_joints, target_pose, method="LM", maxIter=1000):
        """
        Returns joint angles obtained by computing IK
        
        Args:
            current_joints (sequence of float): input joint angles
            target_pose (np.array): goal pose to achieve
            method (str): two methods to calculate IK (LM: Levenberg-marquardt, NR: Newton-raphson)
            maxIter (int): Maximum number of calculation iterations

        Returns:
            joints (np.array): target joint angles
        """
        if not isinstance(target_pose, dict):
            raise TypeError("Be sure to input the target pose in dictionary form.")

        joints = {}
        self._frames = self._input2dict(None)
        self._revolute_joint_names = self._input2dict(None)
        for arm in target_pose.keys():
            if self.eef_name[arm]:
                self._set_desired_frame(arm)
                self._target_pose[arm] = self._convert_target_pose_type_to_npy(target_pose[arm])

                joints[arm] = self.kin.inverse_kinematics(
                    self._frames[arm],
                    current_joints,
                    self._target_pose[arm],
                    method,
                    maxIter)
        return joints

    def _convert_target_pose_type_to_npy(self, value):
        """
        convert input type to numpy array

        Args:
            value(list or tupe)
        
        Returns:
            np.array
        """
        if isinstance(value, (list, tuple)):
            value = np.array(value)
        return value.flatten()

    def get_eef_pose(self, transformations):
        """
        Compute end effector's pose

        Args:
            transformations(OrderedDict)
        
        Returns:
            vals(dict)
        """
        vals = {}
        for arm in self.arm_type:
            if self.eef_name[arm]:
                vals[arm] = np.concatenate((transformations[self.eef_name[arm]].pos, transformations[self.eef_name[arm]].rot))
        return vals

    @property
    def _arms(self):
        """
        Returns name of arms used as naming convention throughout this module

        Returns:
            2-tuple: ('right', 'left')
        """
        return ("right", "left")

    @property
    def arm_type(self):
        """
        Return arm type 
        If number of eef_name is two, return tuple type("right", "left)
        otherwise, return list type(["right] or ["left"])

        Returns:
            arm types (tuple or list)
        """
        if len(self._eef_name.keys()) == 2:
            return self._arms
        elif "right" in self.eef_name.keys():
            return ["right"]
        elif "left" in self.eef_name.keys():
            return ["left"]
        else:
            raise NotFoundError("Can not find robot's arm type")
        
    @property
    def base_name(self):
        return self._base_name

    @property
    def eef_name(self):
        return self._eef_name

    @property
    def active_joint_names(self):
        return self._revolute_joint_names
        



import numpy as np
import time

from pykin.kinematics.transform import Transform

JOINT_TYPE_MAP = {'revolute'  : 'revolute',
                  'fixed'     : 'fixed',
                  'prismatic' : 'prismatic'}

LINK_TYPE_MAP = {'cylinder' : 'cylinder',
                 'sphere'   : 'sphere',
                 'box'      : 'box',
                 'mesh'     : 'mesh'}

LINK_TYPES = ['box', 'cylinder', 'sphere', 'capsule', 'mesh']

class ShellColors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class Baxter:
    left_e0_fixed_offset = Transform(rot=[0.5, 0.5, 0.5, 0.5], pos=[0.107, 0.,    0.   ])
    left_w0_fixed_offset = Transform(rot=[0.5, 0.5, 0.5, 0.5], pos=[0.088, 0.,    0.   ])
    right_e0_fixed_offset = Transform(rot=[0.5, 0.5, 0.5, 0.5], pos=[0.107, 0.,    0.   ])
    right_w0_fixed_offset = Transform(rot=[0.5, 0.5, 0.5, 0.5], pos=[0.088, 0.,    0.   ])

    @staticmethod
    def add_visual_link(link_transforms, f):
        if "left_lower_shoulder" in f.link.name:
            link_transforms["left_upper_elbow_visual"] = np.dot(link_transforms["left_lower_shoulder"],
                                                                        Baxter.left_e0_fixed_offset)
        if "left_lower_elbow" in f.link.name:
            link_transforms["left_upper_forearm_visual"] = np.dot(link_transforms["left_lower_elbow"],
                                                                        Baxter.left_w0_fixed_offset)
        if "right_lower_shoulder" in f.link.name:
            link_transforms["right_upper_elbow_visual"] = np.dot(link_transforms["right_lower_shoulder"],
                                                                        Baxter.right_e0_fixed_offset)
        if "right_lower_elbow" in f.link.name:
            link_transforms["right_upper_forearm_visual"] = np.dot(link_transforms["right_lower_elbow"], 
                                                                        Baxter.right_w0_fixed_offset)


def convert_thetas_to_dict(active_joint_names, thetas):
    """
    Check if any pair of objects in the manager collide with one another.
    
    Args:
        active_joint_names (list): actuated joint names
        thetas (sequence of float): If not dict, convert to dict ex. {joint names : thetas}
    
    Returns:
        thetas (dict): Dictionary of actuated joint angles
    """
    if not isinstance(thetas, dict):
        assert len(active_joint_names) == len(thetas
        ), f"""the number of robot joint's angle is {len(active_joint_names)}, 
                but the number of input joint's angle is {len(thetas)}"""
        thetas = dict((j, thetas[i]) for i, j in enumerate(active_joint_names))        
    return thetas


def logging_time(original_fn):
    """
    Decorator to check time of function
    """
    def wrapper_fn(*args, **kwargs):
        start_time = time.time()
        result = original_fn(*args, **kwargs)
        end_time = time.time()
        print(f"WorkingTime[{original_fn.__name__}]: {end_time-start_time:.4f} sec\n")
        return result
    return wrapper_fn


def convert_transform(origin):
    """
    Args:
        origin (None or Transform): offset of object

    Returns:
        Transform: Returns Transform if origin is None
    """
    if origin is None:
        return Transform()
    else:
        return Transform(rot=origin.rot, pos=origin.pos)


def convert_string_to_narray(str_input):
    """
    Args:
        str_input (str): string

    Returns:
        np.array: Returns string to np.array
    """
    if str_input is not None:
        return np.array([float(data) for data in str_input.split()])


def calc_pose_error(tar_pose, cur_pose, EPS):
    """
    Args:
        tar_pos (np.array): target pose
        cur_pos (np.array): current pose
        EPS (float): epsilon

    Returns:
        np.array: Returns pose error
    """

    pos_err = np.array([tar_pose[:3, -1] - cur_pose[:3, -1]])
    rot_err = np.dot(cur_pose[:3, :3].T, tar_pose[:3, :3])
    w_err = np.dot(cur_pose[:3, :3], rot_to_omega(rot_err, EPS))

    return np.vstack((pos_err.T, w_err))


def rot_to_omega(R, EPS):
    # referred p36
    el = np.array(
            [[R[2, 1] - R[1, 2]],
            [R[0, 2] - R[2, 0]],
            [R[1, 0] - R[0, 1]]]
    )
    norm_el = np.linalg.norm(el)
    if norm_el > EPS:
        w = np.dot(np.arctan2(norm_el, np.trace(R) - 1) / norm_el, el)
    elif (R[0, 0] > 0 and R[1, 1] > 0 and R[2, 2] > 0):
        w = np.zeros((3, 1))
    else:
        w = np.dot(np.pi/2, np.array([[R[0, 0] + 1], [R[1, 1] + 1], [R[2, 2] + 1]]))
    return w


def limit_joints(joint_angles, lower, upper):
    """
    Set joint angle limit

    Args:
        joint_angles (sequence of float): joint angles
        lower (sequence of float): lower limit
        upper (sequence of float): upper limit

    Returns:
        joint_angles (sequence of float): Returns limited joint angle 
    """
    if lower is not None and upper is not None:
        for i in range(len(joint_angles)):
            if joint_angles[i] < lower[i]:
                joint_angles[i] = lower[i]
            if joint_angles[i] > upper[i]:
                joint_angles[i] = upper[i]
    return joint_angles




import io, os
from xml.etree import ElementTree as ET
from collections import OrderedDict

from pykin.models.robot_model import RobotModel
from pykin.geometry.frame import Joint, Link, Frame
from pykin.geometry.geometry import Visual, Collision
from pykin.kinematics.transform import Transform
from pykin.utils.kin_utils import convert_transform, JOINT_TYPE_MAP
from pykin.models.urdf_joint import URDF_Joint
from pykin.models.urdf_link import URDF_Link

class URDFModel(RobotModel):
    """
    Initializes a urdf model, as defined by a single corresponding robot URDF

    Args:
        fname (str): path to the urdf file.
    """
    def __init__(self, fname):
        super().__init__()

        if not os.path.isfile(fname):
            raise FileNotFoundError(f'{fname} is not Found..')

        self.tree_xml = ET.parse(fname)
        self.root = self.tree_xml.getroot()
        self.robot_name = self.root.attrib.get('name')

        self._set_links()
        self._set_joints()
        self._set_root()

    def get_urdf(self):
        """
        Reads a string of the urdf file.

        Returns:
            str: xml read in from file
        """
        with io.StringIO() as string:
            string.write(ET.tostring(self.tree_xml.getroot(), encoding="unicode"))
            return string.getvalue()

    def find_frame(self, frame_name):
        """
        Args:
            frame_name (str): frame's name

        Returns:
            Frame: frame with child frames
        """
        if self.root.name == frame_name:
            return self.root
        return self._find_name_recursive(frame_name, self.root, frame_type="frame")

    def find_link(self, link_name):
        """
        Args:
            link_name (str): link's name

        Returns:
            Link: desired robot's link
        """
        if self.root.link.name == link_name:
            return self.root.link
        return self._find_name_recursive(link_name, self.root, frame_type="link")

    def find_joint(self, joint_name):
        """
        Args:
            joint_name (str): joint's name

        Returns:
            Joint: desired robot's joint
        """
        if self.root.joint.name == joint_name:
            return self.root.joint
        return self._find_name_recursive(joint_name, self.root, frame_type="joint")

    def get_all_active_joint_names(self):
        """
        Returns actuated(revolute, prismatic) joint names

        Args:
            desired_frames (list): If is not empty, will get desired actuated joint names

        Returns:
            list: actuated joint names
        """

        joint_names = []
        if self.root is not None:
            joint_names = self._get_all_active_joint_names_recursive(joint_names, self.root)

        for i, joint in enumerate(joint_names):
            if "head" in joint:
                head_joint = joint_names.pop(i)
                joint_names.insert(0, head_joint)

        return joint_names

    @staticmethod
    def _get_all_active_joint_names_recursive(joint_names, frame):
        """
        Return the name of all actuated joint(revolute, prismatic)

        Args:
            joint_names (list): all actuated joint names
            root_frame (Frame): root frame

        Returns:
            list: Append joint if joint's dof is not zero
        """
        if frame.joint.num_dof != 0:
            joint_names.append(frame.joint.name)
        for child in frame.children:
            URDFModel._get_all_active_joint_names_recursive(joint_names, child)
        return joint_names

    def get_revolute_joint_names(self, desired_frames=None):
        """
        Returns revolute joint names

        Args:
            desired_frames (list): If is not empty, will get desired actuated joint names

        Returns:
            list: revolute joint names
        """
        if desired_frames is None:
            joint_names = self._get_revolute_joint_names(self.root)
        else:
            joint_names = self._get_revolute_joint_names(desired_frames)

        for i, joint in enumerate(joint_names):
            if "head" in joint:
                head_joint = joint_names.pop(i)
                joint_names.insert(0, head_joint)

        return joint_names

    def _set_links(self):
        """
        Set all links from urdf file
        """
        for idx, elem_link in enumerate(self.root.findall('link')):
            link_frame = self._get_link_frame(idx, elem_link)
            self._links[link_frame.link.name] = link_frame.link

    def _set_joints(self):
        """
        Set all joints from urdf file
        """
        for idx, elem_joint in enumerate(self.root.findall('joint')):
            joint_frame = self._get_joint_frame(idx, elem_joint)
            self._joints[joint_frame.joint.name] = joint_frame.joint

    def _set_root(self):
        """
        Set root from urdf file
        """
        root_name = next(iter(self._links))
        self._root_link = self._links[root_name]

        root_frame = self._generate_root_frame(root_name)
        self.root = root_frame

    def _generate_root_frame(self, root_name):
        """
        Generates root frame with all child frames

        Args:
            root_name (str): root name

        Returns:
            Frame: root frame with all child frames
        """
        root_frame = Frame(root_name + "_frame")
        root_frame.joint = Joint()
        root_frame.link = Link(root_name)
        root_frame.children = self._generate_children_recursive(self._root_link, self._links, self._joints)
        return root_frame

    def _get_link_frame(self, idx, elem_link):
        """
        Returns link frame from urdf file

        Args:
            idx (int): index of link parsed from urdf file
            elem_link (xml.etree.ElementTree.Element): element of link parsed from urdf file

        Returns:
            Frame: link frame with all child frames
        """
        attrib = elem_link.attrib
        link_name = attrib.get('name', 'link_' + str(idx))
        link_frame = Frame(name=link_name + '_frame',
                           link=Link(
                           name=link_name, 
                           offset=Transform(), 
                           visual=Visual(), 
                           collision=Collision()))

        URDF_Link.set_visual(elem_link, link_frame)
        URDF_Link.set_collision(elem_link, link_frame)
        
        return link_frame

    def _get_joint_frame(self, idx, elem_joint):
        """
        Returns joint frame from urdf file

        Args:
            idx (int): index of joint parsed from urdf file
            elem_joint (xml.etree.ElementTree.Element): element of joint parsed from urdf file

        Returns:
            Frame: joint frame with all child frames
        """
        attrib = elem_joint.attrib
        joint_name = attrib.get('name', 'joint_' + str(idx))
        joint_frame = Frame(name=joint_name + '_frame',
                            joint=Joint(
                            name=joint_name, 
                            offset=Transform(), 
                            dtype=attrib['type'], 
                            limit=[None, None]))

        parent_tag = elem_joint.find('parent')
        joint_frame.joint.parent = parent_tag.attrib['link']

        child_tag = elem_joint.find('child')
        joint_frame.joint.child = child_tag.attrib['link']

        URDF_Joint.set_origin(elem_joint, joint_frame)
        URDF_Joint.set_axis(elem_joint, joint_frame)
        URDF_Joint.set_limit(elem_joint, joint_frame)

        return joint_frame
    
    @staticmethod
    def _generate_children_recursive(root_link: Link, links: OrderedDict, joints: OrderedDict) -> list:
        """
        Generates child frame recursive from current joint

        Args:
            root (Link): 
            links (OrderedDict): element of joint parsed from urdf file
            joints (OrderedDict): element of joint parsed from urdf file

        Returns:
            list: Append list If current joint's parent link is root link
        """
        children = []
        for joint in joints.values():
            if joint.parent == root_link.name:
                child_frame = Frame(joint.child + "_frame")
                child_frame.joint = Joint(joint.name, 
                                        offset=convert_transform(joint.offset), 
                                        dtype=JOINT_TYPE_MAP[joint.dtype], 
                                        axis=joint.axis, 
                                        limit=joint.limit)

                child_link = links[joint.child]
                child_frame.link = Link(child_link.name, 
                                        offset=convert_transform(child_link.offset),
                                        visual=child_link.visual,
                                        collision=child_link.collision)

                child_frame.children = URDFModel._generate_children_recursive(child_frame.link, links, joints)
                children.append(child_frame)

        return children

    @staticmethod
    def _find_name_recursive(name, frame, frame_type):
        """
        Return the name of the frame, link, or joint you want to find.

        Args:
            name (str): name you want to find
            frame (Frame): frame from root until it finds the desired name
            frame_type (str): 3 frame types, frame or link or joint

        Returns:
            3 types: Frame, Link, Joint
        """
        for frame in frame.children:
            if frame_type == "frame" and frame.name == name:
                return frame
            if frame_type == "link" and frame.link.name == name:
                return frame.link
            if frame_type == "joint" and frame.joint.name == name:
                return frame.joint
            ret = URDFModel._find_name_recursive(name, frame, frame_type)

            assert (ret != None), f"Not Found {name}, please check the name again"
            return ret

    def _get_revolute_joint_names(self, frame):
        """
        Return the name of the actuated joint(revolute, prismatic)

        Args:
            root_frame (str): root frame
            desired_frames (Frame): frames from root until it finds the desired name

        Returns:
            list: Append joint if joint's dof is not zero
        """
        if not isinstance(frame, list):
            joint_names = []
            joint_names =  self._get_all_revolute_joint_names_recursive(joint_names, frame)
        else:
            joint_names = self._get_desired_revolute_joint_names(frame)

        return joint_names

    @staticmethod
    def _get_all_revolute_joint_names_recursive(joint_names, root_frame):
        """
        Return the name of all revolute joint

        Args:
            joint_names (list): all actuated joint names
            root_frame (Frame): root frame

        Returns:
            list: Append joint if joint's dof is not zero
        """
        if root_frame.joint.dtype == 'revolute':
            joint_names.append(root_frame.joint.name)
        for child in root_frame.children:
            URDFModel._get_all_revolute_joint_names_recursive(joint_names, child)
        return joint_names

    @staticmethod
    def _get_desired_revolute_joint_names(desired_frames):
        """
        Return the name of desired actuated joint(revolute, prismatic)

        Args:
            desired_frames (list): desired actuated joint names

        Returns:
            list: Append joint if joint's dof is not zero
        """
        joint_names = []
        for f in desired_frames:
            if f.joint.dtype == 'revolute':
                joint_names.append(f.joint.name)
        return joint_names

    @property
    def dof(self):
        """
        Returns:
            int: robot's dof
        """
        return sum([joint.num_dof for joint in self.joints.values()])

    @property
    def num_links(self):
        """
        Returns:
            int: number of links
        """
        return len(self.links)

    @property
    def num_joints(self):
        """
        Returns:
            int: number of joints
        """
        return len(self.joints)

    @property
    def num_fixed_joints(self):
        """
        Returns:
            int: number of fixed joints
        """
        return sum([1 for joint in self.joints.values() if joint.num_dof == 0])

    @property
    def num_actuated_joints(self):
        """
        Returns:
            int: number of actuated joints
        """
        return sum([1 for joint in self.joints.values() if joint.num_dof != 0])

    @property
    def num_revolute_joints(self):
        """
        Returns:
            int: number of actuated joints
        """
        return len(self.get_revolute_joint_names())

    @staticmethod
    def generate_desired_frame_recursive(base_frame, eef_name):
        """
        Return frames from base_frame to eef_frame you want to find

        Args:
            base_frame (list): reference frame
            eef_name (str): end effector name

        Returns:
            list: Append frame until child link name is eef name
        """
        for child in base_frame.children:
            if child.link.name == eef_name:
                return [child]
            else:
                frames = URDFModel.generate_desired_frame_recursive(child, eef_name)
                if frames is not None:
                    return [child] + frames



import sys, os
import numpy as np
from collections import defaultdict

pykin_path = os.path.abspath(os.path.dirname(__file__)+"../../" )
sys.path.append(pykin_path)
from pykin.kinematics.transform import Transform

from pykin.utils.error_utils import NotFoundError

class Obstacle():
    """
    Obstacle class 
    Obstacles are noe of three types(sphere, box, cylinder)

    """
    obstacle_types = ["mesh", "sphere", "box", "cylinder"]
    def __init__(self):
        self._obstacles = defaultdict(tuple)
        self._gtype = None

    def __call__(self, *args, **kwards):
        self.add_obstacles(*args, **kwards)

    def __repr__(self):
        return f"{self._obstacles}"

    def __iter__(self):
        items = list(self._obstacles.items())
        items.sort(key=lambda x : (x[1][0], x[0]))
        for key, value in items:
            yield (key, value)
            
    def add_obstacles(
        self, 
        name=None, 
        gtype=None, 
        gparam=None, 
        transform=Transform()):
        """
        Add obstacles

        Args:
            name (str): An identifier for the object
            gtype (str): object type (cylinder, sphere, box)
            gparam (float or tuple): object parameter (radius, length, size)
            transform (np.array): Homogeneous transform matrix for the object
        """
        obs_name = self._convert_name(name)
        self._check_gtype(gtype)
        self._check_gparam(gtype, gparam)
        self.obstacles[obs_name] = (gtype, gparam, transform)

    @staticmethod
    def _convert_name(name):
        """
        convert input name to obstacle name

        Args:
            nam (str): An identifier for the object

        Returns:
            name(str) : obstacles_ + name
        """
        if name and "obstacle" not in name:
            name = "obstacle_" + name
        return name
    
    @staticmethod
    def _check_gtype(gtype):
        """
        check obstacle's geom type
        """
        if gtype not in Obstacle.obstacle_types:
            raise NotFoundError(f"'{gtype}' is not in {Obstacle.obstacle_types}")
    
    @staticmethod
    def _check_gparam(gtype, gparam):
        """
        check obstacle's geom param 
        """
        if not isinstance(gparam, (tuple, list, np.ndarray)):
            gparam = [gparam]
        if gtype == "sphere":
            assert len(gparam) == 1, f"{gtype}'s parameter need only 'radius'"
        if gtype == "box":
            assert len(gparam) == 3, f"{gtype}'s parameter need box 'size(x, y, z)'"
        if gtype == "cylinder":
            assert len(gparam) == 2, f"{gtype}'s parameter need 'radius' and 'length'"

    @property
    def obstacles(self):
        return self._obstacles
    



from pykin.utils.kin_utils import convert_string_to_narray

class URDF_Joint:
    """
    Class of parsing joint info described in URDF
    """
    @staticmethod
    def set_origin(elem_joint, frame):
        """
        Set joint's origin
        """
        elem_origin = elem_joint.find('origin')
        if elem_origin is not None:
            frame.joint.offset.pos = convert_string_to_narray(
                elem_origin.attrib.get('xyz'))
            frame.joint.offset.rot = convert_string_to_narray(
                elem_origin.attrib.get('rpy'))
    
    @staticmethod
    def set_axis(elem_joint, frame):
        """
        Set joint's axis
        """
        elem_axis = elem_joint.find('axis')
        if elem_axis is not None:
            frame.joint.axis = convert_string_to_narray(
                elem_axis.attrib.get('xyz'))

    @staticmethod
    def set_limit(elem_joint, frame):
        """
        Set joint's limit
        """
        elem_limit = elem_joint.find('limit')
        if elem_limit is not None:
            if "lower" in elem_limit.attrib:
                frame.joint.limit[0] = float(elem_limit.attrib["lower"])
            if "upper" in elem_limit.attrib:
                frame.joint.limit[1] = float(elem_limit.attrib["upper"])


from setuptools import setup, find_packages
print(find_packages())
# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    lines = f.readlines()

# remove images from README
lines = [x for x in lines if '.png' not in x]
long_description = ''.join(lines)

setup(
    name="pykin",
    packages=find_packages(exclude=["tests"]),
    install_requires=[
        "numpy",
        "matplotlib",
        "trimesh[easy]",
        "tqdm",
        "pyyaml",
        "python-fcl"
    ],
    eager_resources=['*'],
    include_package_data=True,
    python_requires='>=3',
    description="Robotics Kinematics Library",
    author="Dae Jong Jin",
    url="https://github.com/jdj2261/pykin.git",
	download_url="https://github.com/jdj2261/pykin/archive/refs/heads/main.zip",
    author_email="wlseoeo@gmail.com",
    version="0.4.15",
    long_description=long_description,
    long_description_content_type='text/markdown'
)

# python setup.py sdist bdist_wheel
# twine upload --skip-existing dist/*



class ContactData:
    """
    Data structure for holding information about a collision contact.

    Args:
        names (str): The names of the two objects in order.
        contact (fcl.Contact): The contact in question.
    """

    def __init__(self, names, contact):
        self.names = set(names)
        self._inds = {
            names[0]: contact.b1,
            names[1]: contact.b2
        }
        self._point = contact.pos
        self._depth = contact.penetration_depth

    def __repr__(self):
        return 'pykin.collision.collision_manager.{}()'.format(type(self).__name__)

    @property
    def point(self):
        """
        The 3D point of intersection for this contact.
        
        Returns:
            (3,) float: The intersection point.
        """
        return self._point

    @property
    def depth(self):
        """
        The penetration depth of the 3D point of intersection for this contact.
        
        Returns:
            float: The penetration depth.
        """
        return self._depth

    def index(self, name):
        """
        Returns the index of the face in contact for the mesh with
        the given name.

        Args:
            name (str): The name of the target object.

        Returns:
            int: The index of the face in collision
        """
        return self._inds[name]



import numpy as np

from pykin.kinematics.transform import Transform
from pykin.geometry.geometry import Visual, Collision
from pykin.utils import transform_utils as tf
from pykin.utils.kin_utils import ShellColors as scolors


class Link:
    """class of Link

    Args:
        name (str): link name
        offset (pykin.kinematics.transform.Transform): link offset described in the urdf file
        visual (pykin.geometry.geometry.Visual): link visual described in the urdf file
        collision (pykin.geometry.geometry.Collision): link collision described in the urdf file
    """
    def __init__(
        self, 
        name=None, 
        offset=Transform(), 
        visual=Visual(), 
        collision=Collision()
    ):
        self.name = name
        self.offset = offset
        self.visual = visual
        self.collision = collision

    def __str__(self):
        return f"""
        {scolors.OKBLUE}Link{scolors.ENDC}( name= {scolors.HEADER}{self.name}{scolors.ENDC}
            offset= {scolors.HEADER}{self.offset}{scolors.ENDC}
            visual= {scolors.HEADER}{self.visual}{scolors.ENDC} 
            collision= {scolors.HEADER}{self.collision}{scolors.ENDC}"""
    
    def __repr__(self):
        return 'pykin.geometry.frame.{}()'.format(type(self).__name__)
        
class Joint:
    """
    class of Joint

    Args:
        name (str): join name
        offset (pykin.kinematics.transform.Transform): joint offset described in the urdf file
        dtype (str): joint type (fixed, revolute, prismatic) described in the urdf file
        axis (np.array): joint axis described in the urdf file
        limit (list): joint limit described in the urdf file
        parent (Link): joint parent link described in the urdf file
        child (Link): joint child link described in the urdf file
    """
    TYPES = ['fixed', 'revolute', 'prismatic']

    def __init__(
        self,
        name=None, 
        offset=Transform(),
        dtype='fixed', 
        axis=None, 
        limit=[None, None], 
        parent=None, 
        child=None
    ):
        self.name = name
        self.offset = offset
        self.num_dof = 0
        self.dtype = dtype
        self.axis = np.array(axis)
        self.limit = limit
        self.parent = parent
        self.child = child

    def __str__(self):
        return f"""
        {scolors.OKGREEN}Joint{scolors.ENDC}( name= {scolors.HEADER}{self.name}{scolors.ENDC} 
            offset= {scolors.HEADER}{self.offset}{scolors.ENDC}
            dtype= {scolors.HEADER}'{self.dtype}'{scolors.ENDC}
            axis= {scolors.HEADER}{self.axis}{scolors.ENDC}
            limit= {scolors.HEADER}{self.limit}{scolors.ENDC}"""

    def __repr__(self):
        return 'pykin.geometry.frame.{}()'.format(type(self).__name__)

    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, dtype):
        """
        Sets dof 0 if dtype is fixed else 1
        """
        if dtype is not None:
            dtype = dtype.lower().strip()
            if dtype in {'fixed'}:
                dtype = 'fixed'
                self.num_dof = 0
            elif dtype in {'revolute'}:
                dtype = 'revolute'
                self.num_dof = 1
            elif dtype in {'prismatic'}:
                dtype = 'prismatic'
                self.num_dof = 1
        self._dtype = dtype

    @property
    def num_dof(self):
        return self._num_dof

    @num_dof.setter
    def num_dof(self, dof):
        self._num_dof = int(dof)


class Frame:
    """class of Frame

    Args:
        name (str): frame name
        link (Link): Link frame
        joint (Joint): Joint frame
        children (list): all child frame
    """
    def __init__(
        self, 
        name=None, 
        link=Link(),
        joint=Joint(), 
        children=[]
    ):
        self.name = 'None' if name is None else name
        self.link = link
        self.joint = joint
        self.children = children

    def __str__(self, level=0):
        ret = "  " * level + self.name + "\n"
        for child in self.children:
            ret += child.__str__(level + 1)
        return ret

    def __repr__(self):
        return 'pykin.geometry.frame.{}()'.format(type(self).__name__)

    def get_transform(self, theta):
        """
        Args:
            theta (list): Angle to convert

        Returns:
            Transform: Compute transform by multiplying current joint offset and transfrom obtained from input angle
        """
        if self.joint.dtype == 'revolute':
            t = Transform(rot=tf.get_quaternion_about_axis(theta, self.joint.axis))
        elif self.joint.dtype == 'prismatic':
            t = Transform(pos=theta * self.joint.axis)
        elif self.joint.dtype == 'fixed':
            t = Transform()
        else:
            raise ValueError("Unsupported joint type %s." %self.joint.dtype)
        return self.joint.offset * t
